from __future__ import annotations

import logging
import sys
from pprint import pformat as pf
from typing import TYPE_CHECKING, Any, Callable, Final, Generator, Literal

import pandas as pd
import pytest
from packaging.version import Version

import great_expectations.core.batch_spec
from great_expectations.datasource.fluent import BatchRequest
from great_expectations.datasource.fluent.fabric import (
    FabricPowerBIDatasource,
    _PowerBIAsset,
)
from great_expectations.exceptions.exceptions import BuildBatchRequestError

if TYPE_CHECKING:
    from great_expectations.data_context import AbstractDataContext

param = pytest.param

LOGGER = logging.getLogger(__name__)

PYTHON_VERSION: Version = Version(f"{sys.version_info.major}.{sys.version_info.minor}")

_DUMMY_POWER_BI_DATASET_ID: Final[str] = "0993ddd1-f01d-4c8f-a64e-9c4e5b8905e8"


@pytest.fixture
def empty_data_context(empty_data_context: AbstractDataContext) -> AbstractDataContext:
    return empty_data_context


@pytest.fixture
def patch_power_bi_datasource(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Monkeypatch the PowerBI datasource to enable datasource creation even without `semantic-link` installed
    """  # noqa: E501 # FIXME CoP
    monkeypatch.setattr(FabricPowerBIDatasource, "test_connection", lambda _: True)


@pytest.fixture
def power_bi_datasource(
    empty_data_context: AbstractDataContext, patch_power_bi_datasource: None
) -> FabricPowerBIDatasource:
    datasource = empty_data_context.data_sources.add_fabric_powerbi(
        "my_power_bi_datasource",
        dataset=_DUMMY_POWER_BI_DATASET_ID,
        workspace=None,
    )
    return datasource


@pytest.fixture
def capture_reader_fn_params(
    monkeypatch: pytest.MonkeyPatch,
) -> Generator[tuple[list[tuple], list[dict]], None, None]:
    """
    Capture the `reader_options` arguments being passed to the `PandasExecutionEngine`.

    Note this fixture is heavily reliant on the implementation details of `PandasExecutionEngine`,
    should this change this fixture will need to change.
    """
    captured_args: list[tuple] = []
    captured_kwargs: list[dict[str, Any]] = []

    def reader_fn_spy(*args, **kwargs) -> pd.DataFrame:
        LOGGER.info(f"reader_fn_spy() called with...\n{args}\n{kwargs}")
        captured_args.append(args)
        captured_kwargs.append(kwargs)
        return pd.DataFrame.from_dict({"foo": ["bar"]})

    monkeypatch.setattr(
        great_expectations.core.batch_spec.FabricBatchSpec,
        "get_reader_function",
        lambda _: reader_fn_spy,
        raising=True,
    )
    LOGGER.info("Patcehd core.batch_spec.FabricBatchSpec.reader_method")

    yield captured_args, captured_kwargs


@pytest.mark.unit
class TestFabricPowerBI:
    @pytest.mark.parametrize(
        ["asset_type", "asset_kwargs"],
        [
            param("powerbi_dax", {"dax_string": "my_dax_string"}, id="dax min_args"),
            param(
                "powerbi_measure",
                {
                    "measure": "my_measure",
                },
                id="measure min_args",
            ),
            param(
                "powerbi_measure",
                {
                    "measure": "my_measure",
                    "groupby_columns": ["foo[Bar]", "'fizz with space'[Buzz]"],
                    "filters": {
                        "State[Region]": ["East", "Central"],
                        "State[State]": ["WA", "CA"],
                    },
                    "fully_qualified_columns": True,
                    "num_rows": 100,
                    "use_xmla": True,
                },
                id="measure all_args",
            ),
            param("powerbi_table", {"table": "my_table_name"}, id="table min_args"),
        ],
    )
    def test_reader_options_passthrough(
        self,
        power_bi_datasource: FabricPowerBIDatasource,
        capture_reader_fn_params: tuple[list[tuple], list[dict]],
        asset_type: Literal["powerbi_dax", "powerbi_measure", "powerbi_table"],
        asset_kwargs: dict,
    ):
        """Test that the `reader_options` are passed through to the `PandasExecutionEngine`"""
        print(power_bi_datasource)
        # intitial datasource should have no assets
        assert not power_bi_datasource.assets

        add_asset_fn: Callable[[str], _PowerBIAsset] = getattr(
            power_bi_datasource, f"add_{asset_type}_asset"
        )
        my_asset = add_asset_fn(f"my_{asset_type}_asset", **asset_kwargs)
        batch_request = my_asset.build_batch_request()
        my_asset.get_batch(batch_request)

        _, captured_kwargs = capture_reader_fn_params
        print(f"keyword args:\n{pf(captured_kwargs[-1])}")
        for asset_kwarg in asset_kwargs:
            assert asset_kwarg in captured_kwargs[-1]

    @pytest.mark.parametrize(
        ["asset_type", "asset_kwargs"],
        [
            param("powerbi_dax", {"dax_string": "my_dax_string"}, id="dax min_args"),
            param(
                "powerbi_measure",
                {
                    "measure": "my_measure",
                },
                id="measure min_args",
            ),
            param(
                "powerbi_measure",
                {
                    "measure": "my_measure",
                    "groupby_columns": ["foo[Bar]", "'fizz with space'[Buzz]"],
                    "filters": {
                        "State[Region]": ["East", "Central"],
                        "State[State]": ["WA", "CA"],
                    },
                    "fully_qualified_columns": True,
                    "num_rows": 100,
                    "use_xmla": True,
                },
                id="measure all_args",
            ),
            param("powerbi_table", {"table": "my_table_name"}, id="table min_args"),
        ],
    )
    def test_valid_build_batch_request_params(
        self,
        power_bi_datasource: FabricPowerBIDatasource,
        asset_type: Literal["powerbi_dax", "powerbi_measure", "powerbi_table"],
        asset_kwargs: dict,
    ):
        """
        Test that passing valid keyword parameters to `build_batch_request` does not error.

        This test is a bit redundant since mypy should enforce that the method signature is correct,
        but we experienced an error here as part of the initial 1.0 release.
        """
        add_asset_fn: Callable[[str], _PowerBIAsset] = getattr(
            power_bi_datasource, f"add_{asset_type}_asset"
        )
        my_asset = add_asset_fn(f"my_{asset_type}_asset", **asset_kwargs)

        batch_request = my_asset.build_batch_request(
            options=None, batch_slice=None, partitioner=None
        )
        assert isinstance(batch_request, BatchRequest)

    def test_invalid_build_batch_request_params(self, power_bi_datasource: FabricPowerBIDatasource):
        """Test that passing invalid parameters to `build_batch_request` raises an error"""
        with pytest.raises(BuildBatchRequestError):
            power_bi_datasource.add_powerbi_dax_asset(
                "my_dax_asset", dax_string="my_dax_string"
            ).build_batch_request(options={"not": "valid"})


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar, Tuple, Type

from great_expectations.compatibility.typing_extensions import override
from great_expectations.exceptions import (
    BatchDefinitionNotAddedError,
    CheckpointNotAddedError,
    CheckpointRelatedResourcesFreshnessError,
    ExpectationSuiteNotAddedError,
    GreatExpectationsError,
    ResourceFreshnessAggregateError,
    ValidationDefinitionNotAddedError,
    ValidationDefinitionRelatedResourcesFreshnessError,
)


@dataclass
class FreshnessDiagnostics:
    """
    Wrapper around a list of errors; used to determine if a resource has been added successfully
    and is "fresh" or up-to-date with its persisted equivalent.

    Note that some resources may have dependencies on other resources - in order to be considered
    "fresh", the root resource and all of its dependencies must be "fresh".
    For example, a Checkpoint may have dependencies on ValidationDefinitions, which may have
    dependencies on ExpectationSuites and BatchDefinitions.

    GX requires that all resources are persisted successfully before they can be used to prevent
    unexpected behavior.
    """

    raise_for_error_class: ClassVar[Type[ResourceFreshnessAggregateError]] = (
        ResourceFreshnessAggregateError
    )
    errors: list[GreatExpectationsError]

    @property
    def success(self) -> bool:
        return len(self.errors) == 0

    def raise_for_error(self) -> None:
        """
        Conditionally raises an error if the resource has not been added successfully;
        should prescribe the correct action(s) to take.
        """
        if not self.success:
            raise self.raise_for_error_class(errors=self.errors)


@dataclass
class BatchDefinitionFreshnessDiagnostics(FreshnessDiagnostics):
    pass


@dataclass
class ExpectationSuiteFreshnessDiagnostics(FreshnessDiagnostics):
    pass


@dataclass
class _ParentFreshnessDiagnostics(FreshnessDiagnostics):
    """
    Freshness diagnostics for a class that has a natural parent/child relationship
    with other classes.

    All errors throughout the hierarchy should be collected in the parent diagnostics object.
    """

    parent_error_class: ClassVar[Type[GreatExpectationsError]]
    children_error_classes: ClassVar[Tuple[Type[GreatExpectationsError], ...]]

    def update_with_children(self, *children_diagnostics: FreshnessDiagnostics) -> None:
        for diagnostics in children_diagnostics:
            # Child errors should be prepended to parent errors so diagnostics are in order
            self.errors = diagnostics.errors + self.errors

    @property
    def parent_added(self) -> bool:
        return all(not isinstance(err, self.parent_error_class) for err in self.errors)

    @property
    def children_added(self) -> bool:
        return all(not isinstance(err, self.children_error_classes) for err in self.errors)

    @override
    def raise_for_error(self) -> None:
        if not self.success:
            raise self.raise_for_error_class(errors=self.errors)


@dataclass
class ValidationDefinitionFreshnessDiagnostics(_ParentFreshnessDiagnostics):
    parent_error_class: ClassVar[Type[GreatExpectationsError]] = ValidationDefinitionNotAddedError
    children_error_classes: ClassVar[Tuple[Type[GreatExpectationsError], ...]] = (
        ExpectationSuiteNotAddedError,
        BatchDefinitionNotAddedError,
    )
    raise_for_error_class: ClassVar[Type[ResourceFreshnessAggregateError]] = (
        ValidationDefinitionRelatedResourcesFreshnessError
    )


@dataclass
class CheckpointFreshnessDiagnostics(_ParentFreshnessDiagnostics):
    parent_error_class: ClassVar[Type[GreatExpectationsError]] = CheckpointNotAddedError
    children_error_classes: ClassVar[Tuple[Type[GreatExpectationsError], ...]] = (
        ValidationDefinitionNotAddedError,
    )
    raise_for_error_class: ClassVar[Type[ResourceFreshnessAggregateError]] = (
        CheckpointRelatedResourcesFreshnessError
    )



from __future__ import annotations

import logging
from typing import Any

from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from great_expectations.expectations.metrics.column_aggregate_metric_provider import (
    ColumnAggregateMetricProvider,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.expectations.metrics.util import get_dialect_regex_expression

logger = logging.getLogger(__name__)


class ColumnValuesMatchRegexValues(ColumnAggregateMetricProvider):
    metric_name = "column_values.match_regex"

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(  # FIXME CoP
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: dict[str, Any],
        runtime_configuration: dict,
    ) -> list[str]:
        _dialect = execution_engine.dialect_module
        assert _dialect is not None

        (
            selectable,
            _,
            accessor_domain_kwargs,
        ) = execution_engine.get_compute_domain(metric_domain_kwargs, MetricDomainTypes.COLUMN)
        column_name: str = accessor_domain_kwargs["column"]
        column: sa.ColumnClause = sa.column(column_name)
        regex = metric_value_kwargs["regex"]
        limit = metric_value_kwargs["limit"]

        regex_expression = get_dialect_regex_expression(column, regex, _dialect)
        if regex_expression is None:
            logger.warning(f"Regex is not supported for dialect {_dialect.name!s}")
            raise NotImplementedError

        query = sa.select(column).where(regex_expression).select_from(selectable)  # type: ignore[arg-type]
        if isinstance(limit, int):
            query = query.limit(limit)

        rows = execution_engine.execute_query(query).fetchall()

        return [row[0] for row in rows]



from __future__ import annotations

from numbers import Number
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Type, Union

import numpy as np

from great_expectations.compatibility import pydantic
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.exceptions import InvalidExpectationConfigurationError
from great_expectations.expectations.expectation import (
    COLUMN_DESCRIPTION,
    ColumnAggregateExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.render import (
    AtomicDiagnosticRendererType,
    AtomicPrescriptiveRendererType,
    LegacyDescriptiveRendererType,
    LegacyDiagnosticRendererType,
    LegacyRendererType,
    RenderedAtomicContent,
    RenderedStringTemplateContent,
    RenderedTableContent,
    renderedAtomicValueSchema,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererSchema,
    RendererTableValue,
    RendererValueType,
)
from great_expectations.render.util import (
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)
from great_expectations.util import isclose

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs
    from great_expectations.validator.validator import (
        ValidationDependencies,
    )


class QuantileRange(pydantic.BaseModel):
    quantiles: List[float]
    value_ranges: List[List[Union[float, int, None]]]


EXPECTATION_SHORT_DESCRIPTION = (
    "Expect the specific provided column quantiles "
    "to be between a minimum value and a maximum value."
)
QUANTILE_RANGES_DESCRIPTION = (
    "Key 'quantiles' is an increasingly ordered list of desired quantile values (floats). "
    "Key 'value_ranges' is a list of 2-value lists that specify a lower and upper bound "
    "(inclusive) for the corresponding quantile (with [min, max] ordering)."
)
ALLOW_RELATIVE_ERROR_DESCRIPTION = (
    "Whether to allow relative error in quantile "
    "communications on backends that support or require it."
)
DATA_QUALITY_ISSUES = [DataQualityIssues.NUMERIC.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectColumnQuantileValuesToBeBetween(ColumnAggregateExpectation):
    # noinspection PyUnresolvedReferences
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnQuantileValuesToBeBetween is a \
    Column Aggregate Expectation.

    Column Aggregate Expectations are one of the most common types of Expectation.
    They are evaluated for a single column, and produce an aggregate Metric, such as a mean, standard deviation, number of unique values, column type, etc.
    If that Metric meets the conditions you set, the Expectation considers that data valid.

    ExpectColumnQuantileValuesToBeBetween can be computationally intensive for large datasets.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        quantile_ranges (dictionary with keys 'quantiles' and 'value_ranges'): \
            {QUANTILE_RANGES_DESCRIPTION} The length of the 'quantiles' list and the 'value_ranges' list must be equal.
        allow_relative_error (boolean or string): \
            {ALLOW_RELATIVE_ERROR_DESCRIPTION}

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Notes:
        * min_value and max_value are both inclusive.
        * If min_value is None, then max_value is treated as an upper bound only
        * If max_value is None, then min_value is treated as a lower bound only
        * details.success_details field in the result object is customized for this expectation

    See Also:
        [ExpectColumnMinToBeBetween](https://greatexpectations.io/expectations/expect_column_min_to_be_between)
        [ExpectColumnMaxToBeBetween](https://greatexpectations.io/expectations/expect_column_max_to_be_between)
        [ExpectColumnMedianToBeBetween](https://greatexpectations.io/expectations/expect_column_median_to_be_between)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test
            0 	1       1
            1 	2       7
            2 	2       2.5
            3   3       3
            4   3       2
            5   3       5
            6   4       6

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnQuantileValuesToBeBetween(
                    column="test",
                    quantile_ranges={{
                        "quantiles": [0, .333, .667, 1],
                        "value_ranges": [[0,1], [2,3], [3,4], [4,5]]
                    }}
                )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": {{
                      "quantiles": [
                        0,
                        0.333,
                        0.6667,
                        1
                      ],
                      "values": [
                        1,
                        2,
                        3,
                        4
                      ]
                    }},
                    "details": {{
                      "success_details": [
                        true,
                        true,
                        true,
                        true
                      ]
                    }}
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnQuantileValuesToBeBetween(
                    column="test2",
                    quantile_ranges={{
                        "quantiles": [0, .333, .667, 1],
                        "value_ranges": [[0,1], [2,3], [3,4], [4,5]]
                    }}
                )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": {{
                      "quantiles": [
                        0,
                        0.333,
                        0.6667,
                        1
                      ],
                      "values": [
                        1.0,
                        2.5,
                        5.0,
                        7.0
                      ]
                    }},
                    "details": {{
                      "success_details": [
                        true,
                        true,
                        false,
                        false
                      ]
                    }}
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    quantile_ranges: Union[QuantileRange, SuiteParameterDict] = pydantic.Field(
        description=QUANTILE_RANGES_DESCRIPTION
    )
    allow_relative_error: Union[bool, str, SuiteParameterDict] = pydantic.Field(
        default=False,
        description=ALLOW_RELATIVE_ERROR_DESCRIPTION,
    )

    # This dictionary contains metadata for display in the public gallery
    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "column aggregate expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    _library_metadata = library_metadata

    metric_dependencies = ("column.quantile_values",)
    success_keys = (
        "quantile_ranges",
        "allow_relative_error",
    )

    args_keys = (
        "column",
        "quantile_ranges",
        "allow_relative_error",
    )

    class Config:
        title = "Expect column quantile values to be between"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnQuantileValuesToBeBetween]
        ) -> None:
            ColumnAggregateExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @pydantic.validator("quantile_ranges")
    def validate_quantile_ranges(
        cls, quantile_ranges: QuantileRange | SuiteParameterDict
    ) -> Optional[QuantileRange | SuiteParameterDict]:
        if isinstance(quantile_ranges, QuantileRange):
            try:
                assert all(
                    True if None in x else x == sorted([val for val in x if val is not None])
                    for x in quantile_ranges.value_ranges
                ), "quantile_ranges must consist of ordered pairs"
            except AssertionError as e:
                raise InvalidExpectationConfigurationError(str(e))

            if len(quantile_ranges.quantiles) != len(quantile_ranges.value_ranges):
                raise ValueError(  # noqa: TRY003
                    "quantile_values and quantiles must have the same number of elements"
                )  # FIXME CoP

        return quantile_ranges

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        template_str = "quantiles must be within the following value ranges."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        quantiles: list = renderer_configuration.kwargs.get("quantile_ranges", {}).get(
            "quantiles", []
        )
        value_ranges: list = renderer_configuration.kwargs.get("quantile_ranges", {}).get(
            "value_ranges", []
        )

        header_row = [
            RendererTableValue(
                schema=RendererSchema(type=RendererValueType.STRING), value="Quantile"
            ),
            RendererTableValue(
                schema=RendererSchema(type=RendererValueType.STRING), value="Min Value"
            ),
            RendererTableValue(
                schema=RendererSchema(type=RendererValueType.STRING), value="Max Value"
            ),
        ]

        renderer_configuration.header_row = header_row

        table = []
        quantile_strings = {0.25: "Q1", 0.75: "Q3", 0.50: "Median"}
        for quantile, value_range in zip(quantiles, value_ranges):
            quantile_string = quantile_strings.get(quantile, f"{quantile:3.2f}")
            value_range_lower: Union[Number, str] = value_range[0] if value_range[0] else "Any"
            value_rage_lower_type = (
                RendererValueType.NUMBER if value_range[0] else RendererValueType.STRING
            )
            value_range_upper: Union[Number, str] = value_range[1] if value_range[1] else "Any"
            value_range_upper_type = (
                RendererValueType.NUMBER if value_range[0] else RendererValueType.STRING
            )
            table.append(
                [
                    RendererTableValue(
                        schema=RendererSchema(type=RendererValueType.STRING),
                        value=quantile_string,
                    ),
                    RendererTableValue(
                        schema=RendererSchema(type=value_rage_lower_type),
                        value=value_range_lower,
                    ),
                    RendererTableValue(
                        schema=RendererSchema(type=value_range_upper_type),
                        value=value_range_upper,
                    ),
                ]
            )

        renderer_configuration.table = table

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=AtomicPrescriptiveRendererType.SUMMARY)
    @render_suite_parameter_string
    def _prescriptive_summary(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> RenderedAtomicContent:
        renderer_configuration: RendererConfiguration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        renderer_configuration = cls._prescriptive_template(
            renderer_configuration=renderer_configuration
        )
        header_row = [value.dict() for value in renderer_configuration.header_row]
        table = []
        for row in renderer_configuration.table:
            table.append([value.dict() for value in row])
        value_obj = renderedAtomicValueSchema.load(
            {
                "header": {
                    "schema": {"type": "StringValueType"},
                    "value": {
                        "template": renderer_configuration.template_str,
                        "params": renderer_configuration.params.dict(),
                    },
                },
                "header_row": header_row,
                "table": table,
                "meta_notes": renderer_configuration.meta_notes,
                "schema": {"type": "TableType"},
            }
        )
        return RenderedAtomicContent(
            name=AtomicPrescriptiveRendererType.SUMMARY,
            value=value_obj,
            value_type="TableType",
        )

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        _ = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration["kwargs"],
            ["column", "quantile_ranges", "row_condition", "condition_parser"],
        )
        template_str = "quantiles must be within the following value ranges."

        if include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = (
                conditional_template_str + ", then " + template_str[0].lower() + template_str[1:]
            )
            params.update(conditional_params)

        expectation_string_obj = {
            "content_block_type": "string_template",
            "string_template": {"template": template_str, "params": params},
        }

        quantiles = params["quantile_ranges"].get("quantiles", [])
        value_ranges = params["quantile_ranges"].get("value_ranges", [])

        table_header_row = ["Quantile", "Min Value", "Max Value"]
        table_rows = []

        quantile_strings = {0.25: "Q1", 0.75: "Q3", 0.50: "Median"}

        for quantile, value_range in zip(quantiles, value_ranges):
            quantile_string = quantile_strings.get(quantile, f"{quantile:3.2f}")
            table_rows.append(
                [
                    quantile_string,
                    str(value_range[0]) if value_range[0] is not None else "Any",
                    str(value_range[1]) if value_range[1] is not None else "Any",
                ]
            )

        quantile_range_table = RenderedTableContent(
            **{
                "content_block_type": "table",
                "header_row": table_header_row,
                "table": table_rows,
                "styling": {
                    "body": {
                        "classes": [
                            "table",
                            "table-sm",
                            "table-unbordered",
                            "col-4",
                            "mt-2",
                        ],
                    },
                    "parent": {"styles": {"list-style-type": "none"}},
                },
            }
        )

        return [expectation_string_obj, quantile_range_table]

    @classmethod
    @renderer(renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE)
    def _diagnostic_observed_value_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        if result.result is None or result.result.get("observed_value") is None:
            return "--"

        quantiles = result.result.get("observed_value", {}).get("quantiles", [])
        value_ranges = result.result.get("observed_value", {}).get("values", [])

        table_header_row = ["Quantile", "Value"]
        table_rows = []

        quantile_strings = {0.25: "Q1", 0.75: "Q3", 0.50: "Median"}

        for idx, quantile in enumerate(quantiles):
            quantile_string = quantile_strings.get(quantile)
            table_rows.append(
                [
                    quantile_string if quantile_string else f"{quantile:3.2f}",
                    str(value_ranges[idx]),
                ]
            )

        return RenderedTableContent(
            **{
                "content_block_type": "table",
                "header_row": table_header_row,
                "table": table_rows,
                "styling": {
                    "body": {
                        "classes": ["table", "table-sm", "table-unbordered", "col-4"],
                    }
                },
            }
        )

    @classmethod
    def _atomic_diagnostic_observed_value_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ):
        if (
            renderer_configuration.result.result is None
            or renderer_configuration.result.result.get("observed_value") is None
        ):
            renderer_configuration.template_str = "--"
            return renderer_configuration

        quantiles = renderer_configuration.result.result.get("observed_value", {}).get(
            "quantiles", []
        )
        value_ranges = renderer_configuration.result.result.get("observed_value", {}).get(
            "values", []
        )

        header_row = [
            RendererTableValue(
                schema=RendererSchema(type=RendererValueType.STRING), value="Quantile"
            ),
            RendererTableValue(schema=RendererSchema(type=RendererValueType.STRING), value="Value"),
        ]

        table = []
        quantile_strings = {0.25: "Q1", 0.75: "Q3", 0.50: "Median"}
        for idx, quantile in enumerate(quantiles):
            quantile_string = quantile_strings.get(quantile) or f"{quantile:3.2f}"
            table.append(
                [
                    RendererTableValue(
                        schema=RendererSchema(type=RendererValueType.STRING),
                        value=quantile_string,
                    ),
                    RendererTableValue(
                        schema=RendererSchema(type=RendererValueType.NUMBER),
                        value=value_ranges[idx],
                    ),
                ]
            )

        renderer_configuration.header_row = header_row
        renderer_configuration.table = table

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=AtomicDiagnosticRendererType.OBSERVED_VALUE)
    def _atomic_diagnostic_observed_value(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        renderer_configuration: RendererConfiguration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        renderer_configuration = cls._atomic_diagnostic_observed_value_template(
            renderer_configuration=renderer_configuration,
        )
        if renderer_configuration.template_str:
            value_obj = renderedAtomicValueSchema.load(
                {
                    "template": renderer_configuration.template_str,
                    "params": {},
                    "schema": {"type": "StringValueType"},
                }
            )
            return RenderedAtomicContent(
                name="atomic.diagnostic.observed_value",
                value=value_obj,
                value_type="StringValueType",
            )
        else:
            header_row = [value.dict() for value in renderer_configuration.header_row]
            table = []
            for row in renderer_configuration.table:
                table.append([value.dict() for value in row])
            value_obj = renderedAtomicValueSchema.load(
                {
                    "header_row": header_row,
                    "table": table,
                    "schema": {"type": "TableType"},
                }
            )
            return RenderedAtomicContent(
                name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
                value=value_obj,
                value_type="TableType",
            )

    @classmethod
    @renderer(renderer_type=LegacyDescriptiveRendererType.QUANTILE_TABLE)
    def _descriptive_quantile_table_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        assert result, "Must pass in result."
        table_rows = []
        quantiles = result.result["observed_value"]["quantiles"]
        quantile_ranges = result.result["observed_value"]["values"]

        quantile_strings = {0.25: "Q1", 0.75: "Q3", 0.50: "Median"}

        for idx, quantile in enumerate(quantiles):
            quantile_string = quantile_strings.get(quantile)
            table_rows.append(
                [
                    {
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": (
                                quantile_string if quantile_string else f"{quantile:3.2f}"
                            ),
                            "tooltip": {
                                "content": "expect_column_quantile_values_to_be_between \n expect_column_median_to_be_between"  # noqa: E501 # FIXME CoP
                                if quantile == 0.50  # noqa: PLR2004 # FIXME CoP
                                else "expect_column_quantile_values_to_be_between"
                            },
                        },
                    },
                    quantile_ranges[idx],
                ]
            )

        return RenderedTableContent(
            **{
                "content_block_type": "table",
                "header": RenderedStringTemplateContent(
                    **{
                        "content_block_type": "string_template",
                        "string_template": {"template": "Quantiles", "tag": "h6"},
                    }
                ),
                "table": table_rows,
                "styling": {
                    "classes": ["col-3", "mt-1", "pl-1", "pr-1"],
                    "body": {
                        "classes": ["table", "table-sm", "table-unbordered"],
                    },
                },
            }
        )

    def get_validation_dependencies(
        self,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> ValidationDependencies:
        validation_dependencies: ValidationDependencies = super().get_validation_dependencies(
            execution_engine, runtime_configuration
        )
        configuration = self.configuration
        # column.quantile_values expects a "quantiles" key
        validation_dependencies.get_metric_configuration(
            metric_name="column.quantile_values"
        ).metric_value_kwargs["quantiles"] = configuration.kwargs["quantile_ranges"]["quantiles"]
        return validation_dependencies

    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        quantile_vals = metrics.get("column.quantile_values")
        quantile_ranges = self.configuration.kwargs.get("quantile_ranges")
        quantiles = quantile_ranges["quantiles"]
        quantile_value_ranges = quantile_ranges["value_ranges"]

        # We explicitly allow "None" to be interpreted as +/- infinity
        comparison_quantile_ranges = [
            [
                -np.inf if lower_bound is None else lower_bound,
                np.inf if upper_bound is None else upper_bound,
            ]
            for (lower_bound, upper_bound) in quantile_value_ranges
        ]
        success_details = [
            isclose(
                operand_a=quantile_vals[idx],
                operand_b=range_[0],
                rtol=1.0e-4,
            )
            or isclose(
                operand_a=quantile_vals[idx],
                operand_b=range_[1],
                rtol=1.0e-4,
            )
            or range_[0] <= quantile_vals[idx] <= range_[1]
            for idx, range_ in enumerate(comparison_quantile_ranges)
        ]

        return {
            "success": np.all(success_details),
            "result": {
                "observed_value": {"quantiles": quantiles, "values": quantile_vals},
                "details": {"success_details": success_details},
            },
        }



import logging
import os
import re
import uuid
from typing import TYPE_CHECKING, List

import pandas as pd
import pytest
from moto import mock_s3

from great_expectations.core import IDDict
from great_expectations.core.batch import LegacyBatchDefinition
from great_expectations.core.partitioners import FileNamePartitionerPath, FileNamePartitionerYearly
from great_expectations.core.util import S3Url
from great_expectations.datasource.fluent import BatchRequest
from great_expectations.datasource.fluent.data_connector import (
    S3DataConnector,
)
from great_expectations.datasource.fluent.data_connector.azure_blob_storage_data_connector import (
    sanitize_prefix,
)
from great_expectations.datasource.fluent.data_connector.file_path_data_connector import (
    MissingFilePathTemplateMapFnError,
    sanitize_prefix_for_gcs_and_s3,
)

if TYPE_CHECKING:
    from great_expectations.datasource.fluent.data_connector import (
        DataConnector,
    )

logger = logging.getLogger(__name__)

try:
    import boto3  # : disable=E0602
except ImportError:
    logger.debug("Unable to load boto3; install optional boto3 dependency for support.")


@pytest.fixture
def aws_region():
    """AWS region for testing."""
    return "us-east-1"


@pytest.fixture
def s3_bucket_name():
    """S3 bucket name for testing."""
    return f"test-bucket-{uuid.uuid4().hex[:8]}"


@pytest.fixture
def test_df():
    """Sample pandas DataFrame for testing."""
    return pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})


def create_s3_setup(bucket_name: str, region: str = "us-east-1"):
    """Helper function to create S3 resources within a @mock_s3 context."""
    conn = boto3.resource("s3", region_name=region)
    conn.create_bucket(Bucket=bucket_name)
    client = boto3.client("s3", region_name=region)
    return client, bucket_name


@pytest.mark.big
@mock_s3
def test_basic_instantiation(s3_bucket_name, aws_region, test_df):
    client, bucket = create_s3_setup(s3_bucket_name, aws_region)

    keys: List[str] = [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    for key in keys:
        client.put_object(Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key)

    my_data_connector: DataConnector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    batching_regex = re.compile(r"alpha-(.*)\.csv")
    matching_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matching_data_references) == 3
    assert matching_data_references[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]


@pytest.mark.big
@mock_s3
def test_instantiation_batching_regex_does_not_match_paths(s3_bucket_name, aws_region, test_df):
    client, bucket = create_s3_setup(s3_bucket_name, aws_region)

    keys: List[str] = [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    for key in keys:
        client.put_object(Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key)

    my_data_connector: DataConnector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    batching_regex = re.compile(r"beta-(.*)\.csv")
    matching_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matching_data_references) == 0
    assert matching_data_references[:3] == []


@pytest.mark.big
@mock_s3
def test_return_all_batch_definitions_unsorted(s3_bucket_name, aws_region, test_df):
    client, bucket = create_s3_setup(s3_bucket_name, aws_region)

    keys: List[str] = [
        "alex_2020-08-09_1000.csv",
        "eugene_2020-08-09_1500.csv",
        "james_2020-08-11_1009.csv",
        "abe_2020-08-09_1040.csv",
        "will_2020-08-09_1002.csv",
        "james_2020-07-13_1567.csv",
        "eugene_2020-11-29_1900.csv",
        "will_2020-08-10_1001.csv",
        "james_2020-08-10_1003.csv",
        "alex_2020-08-19_1300.csv",
    ]
    for key in keys:
        client.put_object(Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key)
    my_data_connector: DataConnector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    # with missing BatchRequest arguments
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list()

    batching_regex = re.compile(
        r"(?P<name>.+)_(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})_(?P<price>.*)\.csv"
    )
    partitioner = FileNamePartitionerYearly(regex=batching_regex)

    # with empty options
    unsorted_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="my_file_path_datasource",
                data_asset_name="my_s3_data_asset",
                options={},
                partitioner=partitioner,
            )
        )
    )
    processed_batching_regex = re.compile(
        "(?P<path>(?P<name>.+)_(?P<year>\\d{4})-(?P<month>\\d{2})-(?P<day>\\d{2})_(?P<price>.*)\\.csv)"
    )
    expected: List[LegacyBatchDefinition] = [
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "abe_2020-08-09_1040.csv",
                    "name": "abe",
                    "year": "2020",
                    "month": "08",
                    "day": "09",
                    "price": "1040",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "alex_2020-08-09_1000.csv",
                    "name": "alex",
                    "year": "2020",
                    "month": "08",
                    "day": "09",
                    "price": "1000",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "alex_2020-08-19_1300.csv",
                    "name": "alex",
                    "year": "2020",
                    "month": "08",
                    "day": "19",
                    "price": "1300",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "eugene_2020-08-09_1500.csv",
                    "name": "eugene",
                    "year": "2020",
                    "month": "08",
                    "day": "09",
                    "price": "1500",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "eugene_2020-11-29_1900.csv",
                    "name": "eugene",
                    "year": "2020",
                    "month": "11",
                    "day": "29",
                    "price": "1900",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_2020-07-13_1567.csv",
                    "name": "james",
                    "year": "2020",
                    "month": "07",
                    "day": "13",
                    "price": "1567",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_2020-08-10_1003.csv",
                    "name": "james",
                    "year": "2020",
                    "month": "08",
                    "day": "10",
                    "price": "1003",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_2020-08-11_1009.csv",
                    "name": "james",
                    "year": "2020",
                    "month": "08",
                    "day": "11",
                    "price": "1009",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "will_2020-08-09_1002.csv",
                    "name": "will",
                    "year": "2020",
                    "month": "08",
                    "day": "09",
                    "price": "1002",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "will_2020-08-10_1001.csv",
                    "name": "will",
                    "year": "2020",
                    "month": "08",
                    "day": "10",
                    "price": "1001",
                }
            ),
            batching_regex=processed_batching_regex,
        ),
    ]
    assert expected == unsorted_batch_definition_list

    # with specified Batch query options
    unsorted_batch_definition_list = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_file_path_datasource",
            data_asset_name="my_s3_data_asset",
            options={"name": "alex", "year": "2020", "month": "08", "day": "19", "price": "1300"},
            partitioner=partitioner,
        )
    )
    assert expected[2:3] == unsorted_batch_definition_list


@pytest.mark.big
@mock_s3
def test_return_only_unique_batch_definitions(s3_bucket_name, aws_region, test_df):
    client, bucket = create_s3_setup(s3_bucket_name, aws_region)

    keys: List[str] = [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
        "B/file_1.csv",
        "B/file_2.csv",
    ]
    for key in keys:
        client.put_object(Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key)
    processed_batching_regex = re.compile("(?P<path>B/(?P<filename>.+).*\\.csv)")
    expected: List[LegacyBatchDefinition] = [
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict({"path": "B/file_1.csv", "filename": "file_1"}),
            batching_regex=processed_batching_regex,
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict({"path": "B/file_2.csv", "filename": "file_2"}),
            batching_regex=processed_batching_regex,
        ),
    ]

    my_data_connector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="B",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )

    batching_regex = re.compile(r"(?P<filename>.+).*\.csv")
    unsorted_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="my_file_path_datasource",
                data_asset_name="my_s3_data_asset",
                options={},
                partitioner=FileNamePartitionerPath(regex=batching_regex),
            )
        )
    )
    assert expected == unsorted_batch_definition_list


@pytest.mark.big
@mock_s3
def test_data_reference_count_methods(s3_bucket_name, aws_region, test_df):
    client, bucket = create_s3_setup(s3_bucket_name, aws_region)

    keys: List[str] = [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
        "B/file_1.csv",
        "B/file_2.csv",
    ]
    for key in keys:
        client.put_object(Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key)

    my_data_connector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="A",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
    ]

    batching_regex = re.compile(r"(?P<name>.+).*\.csv")
    matching_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matching_data_references) == 3
    assert matching_data_references[:3] == [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
    ]


@pytest.mark.big
@mock_s3
def test_alpha(s3_bucket_name, aws_region, test_df):
    client, bucket = create_s3_setup(s3_bucket_name, aws_region)

    keys: List[str] = [
        "test_dir_alpha/A.csv",
        "test_dir_alpha/B.csv",
        "test_dir_alpha/C.csv",
        "test_dir_alpha/D.csv",
    ]
    for key in keys:
        client.put_object(Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key)

    my_data_connector: DataConnector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="test_dir_alpha",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 4
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_alpha/A.csv",
        "test_dir_alpha/B.csv",
        "test_dir_alpha/C.csv",
    ]

    batching_regex = re.compile(r"(?P<part_1>.*)\.csv")
    matching_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matching_data_references) == 4
    assert matching_data_references[:3] == [
        "test_dir_alpha/A.csv",
        "test_dir_alpha/B.csv",
        "test_dir_alpha/C.csv",
    ]

    # Try to fetch a batch from a nonexistent asset
    my_batch_request = BatchRequest(
        datasource_name="BASE",
        data_asset_name="A",
        options={},
        partitioner=FileNamePartitionerPath(regex=batching_regex),
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 0

    my_batch_request = BatchRequest(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        options={"part_1": "B"},
        partitioner=FileNamePartitionerPath(regex=batching_regex),
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 1


@pytest.mark.big
@mock_s3
def test_foxtrot(s3_bucket_name, aws_region, test_df):
    client, bucket = create_s3_setup(s3_bucket_name, aws_region)

    keys: List[str] = [
        "test_dir_foxtrot/A/A-1.csv",
        "test_dir_foxtrot/A/A-2.csv",
        "test_dir_foxtrot/A/A-3.csv",
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
        "test_dir_foxtrot/C/C-2017.csv",
        "test_dir_foxtrot/C/C-2018.csv",
        "test_dir_foxtrot/C/C-2019.csv",
        "test_dir_foxtrot/D/D-aaa.csv",
        "test_dir_foxtrot/D/D-bbb.csv",
        "test_dir_foxtrot/D/D-ccc.csv",
        "test_dir_foxtrot/D/D-ddd.csv",
        "test_dir_foxtrot/D/D-eee.csv",
    ]
    for key in keys:
        client.put_object(Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key)

    my_data_connector: DataConnector

    my_data_connector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 0
    assert my_data_connector.get_data_references()[:3] == []

    batching_regex = re.compile(r"(?P<part_1>.+)-(?P<part_2>.*)\.csv")
    matching_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matching_data_references) == 0
    assert matching_data_references[:3] == []

    my_data_connector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="test_dir_foxtrot/A",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/A/A-1.csv",
        "test_dir_foxtrot/A/A-2.csv",
        "test_dir_foxtrot/A/A-3.csv",
    ]

    batching_regex = re.compile(r"(?P<part_1>.+)-(?P<part_2>.*)\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 3
    assert matched_data_references[:3] == [
        "test_dir_foxtrot/A/A-1.csv",
        "test_dir_foxtrot/A/A-2.csv",
        "test_dir_foxtrot/A/A-3.csv",
    ]

    my_data_connector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="test_dir_foxtrot/B",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
    ]

    batching_regex = re.compile(r"(?P<part_1>.+)-(?P<part_2>.*)\.txt")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 3
    assert matched_data_references[:3] == [
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
    ]

    my_data_connector = S3DataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        s3_client=client,
        bucket=bucket,
        prefix="test_dir_foxtrot/C",
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/C/C-2017.csv",
        "test_dir_foxtrot/C/C-2018.csv",
        "test_dir_foxtrot/C/C-2019.csv",
    ]

    batching_regex = re.compile(r"(?P<part_1>.+)-(?P<part_2>.*)\.csv")
    matched_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matched_data_references) == 3
    assert matched_data_references[:3] == [
        "test_dir_foxtrot/C/C-2017.csv",
        "test_dir_foxtrot/C/C-2018.csv",
        "test_dir_foxtrot/C/C-2019.csv",
    ]

    my_batch_request = BatchRequest(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_s3_data_asset",
        options={},
    )
    my_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(batch_request=my_batch_request)
    )
    assert len(my_batch_definition_list) == 3


@pytest.mark.unit
def test_sanitize_prefix_behaves_the_same_as_local_files():
    def check_sameness(prefix, expected_output):
        s3_sanitized = sanitize_prefix_for_gcs_and_s3(text=prefix)
        file_system_sanitized = sanitize_prefix(prefix)
        if os.sep == "\\":  # Fix to ensure tests work on Windows
            file_system_sanitized = file_system_sanitized.replace("\\", "/")

        assert file_system_sanitized == expected_output, (
            f"Expected output does not match original sanitization behavior, got "
            f"{file_system_sanitized} instead of {expected_output}"
        )
        assert s3_sanitized == expected_output == file_system_sanitized, (
            f'S3 sanitized result is incorrect, "{s3_sanitized} instead of {expected_output}'
        )

    # Copy of all samples from tests/datasource/data_connector/test_file_path_data_connector.py
    check_sameness("foo/", "foo/")
    check_sameness("bar", "bar/")
    check_sameness("baz.txt", "baz.txt")
    check_sameness("a/b/c/baz.txt", "a/b/c/baz.txt")

    # A couple additional checks
    check_sameness("a/b/c", "a/b/c/")
    check_sameness("a.x/b/c", "a.x/b/c/")
    check_sameness("path/to/folder.something/", "path/to/folder.something/")
    check_sameness("path/to/folder.something", "path/to/folder.something")


@pytest.mark.unit
@mock_s3
@pytest.mark.parametrize(
    "whole_directory_override, expected_batch_count, expected_identifier_key",
    [
        pytest.param(
            True, 1, "path", id="with_whole_directory_override_returns_single_directory_batch"
        ),
        pytest.param(
            False,
            3,
            "filename",
            id="without_whole_directory_override_returns_individual_file_batches",
        ),
    ],
)
def test_s3_data_connector_whole_directory_path_override(
    whole_directory_override,
    expected_batch_count,
    expected_identifier_key,
    s3_bucket_name,
    aws_region,
):
    """Test S3DataConnector behavior with and without whole_directory_path_override parameter."""
    # Setup
    s3_client, bucket_name = create_s3_setup(s3_bucket_name, aws_region)

    prefix = "test_directory/"
    whole_directory_path = f"s3://{bucket_name}/{prefix}"

    # Create multiple files in the directory
    test_files = [
        f"{prefix}file1.csv",
        f"{prefix}file2.csv",
        f"{prefix}file3.csv",
    ]

    for file_path in test_files:
        s3_client.put_object(Bucket=bucket_name, Key=file_path, Body="col1,col2\n1,2\n3,4\n")

    # Create data connector with conditional whole_directory_path_override
    data_connector = S3DataConnector(
        datasource_name="my_s3_datasource",
        data_asset_name="my_data_asset",
        s3_client=s3_client,
        bucket=bucket_name,
        prefix=prefix,
        file_path_template_map_fn=S3Url.OBJECT_URL_TEMPLATE.format,
        whole_directory_path_override=whole_directory_path if whole_directory_override else None,
    )

    # Create batch request with conditional partitioner
    batch_request = BatchRequest(
        datasource_name="my_s3_datasource",
        data_asset_name="my_data_asset",
        options={},
        partitioner=None
        if whole_directory_override
        else FileNamePartitionerPath(regex=re.compile(r"(?P<filename>.+\.csv)")),
    )
    batch_definitions = data_connector.get_batch_definition_list(batch_request)

    # Verify expected batch count
    assert len(batch_definitions) == expected_batch_count

    # Verify batch definitions have correct structure
    for batch_definition in batch_definitions:
        assert batch_definition.datasource_name == "my_s3_datasource"
        assert batch_definition.data_asset_name == "my_data_asset"
        assert expected_identifier_key in batch_definition.batch_identifiers

    if whole_directory_override:
        # For directory mode, verify single batch with directory path
        batch_definition = batch_definitions[0]
        assert batch_definition.batch_identifiers["path"] == whole_directory_path
    else:
        # For file mode, verify individual file batches
        file_names = [bd.batch_identifiers["filename"] for bd in batch_definitions]
        expected_files = ["file1.csv", "file2.csv", "file3.csv"]
        assert sorted(file_names) == sorted(expected_files)


@pytest.mark.unit
@mock_s3
def test_s3_data_connector_missing_file_path_template_map_fn_error(s3_bucket_name, aws_region):
    """Test S3DataConnector raises MissingFilePathTemplateMapFnError
    when missing file_path_template_map_fn."""
    s3_client, bucket = create_s3_setup(s3_bucket_name, aws_region)

    data_connector = S3DataConnector(
        datasource_name="my_s3_datasource",
        data_asset_name="my_data_asset",
        s3_client=s3_client,
        bucket=bucket,
        prefix="test/",
        file_path_template_map_fn=None,
    )

    with pytest.raises(MissingFilePathTemplateMapFnError):
        data_connector._get_full_file_path("test.csv")



from __future__ import annotations

from copy import copy
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from great_expectations import __version__ as ge_version
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
)
from great_expectations.core.result_format import (
    DEFAULT_RESULT_FORMAT,
    ResultFormat,
)
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP
from great_expectations.validator.validator import Validator as OldValidator
from great_expectations.validator.validator import calc_validation_statistics

if TYPE_CHECKING:
    from great_expectations.core import ExpectationSuite
    from great_expectations.core.batch_definition import BatchDefinition
    from great_expectations.core.result_format import ResultFormatUnion
    from great_expectations.core.suite_parameters import SuiteParameterDict
    from great_expectations.datasource.fluent.batch_request import BatchParameters
    from great_expectations.expectations.expectation import (
        Expectation,
        ExpectationConfiguration,
    )


class Validator:
    """Validator.

    Responsible for running expectations on a batch definition.
    """

    def __init__(
        self,
        batch_definition: BatchDefinition,
        result_format: ResultFormatUnion = DEFAULT_RESULT_FORMAT,
        batch_parameters: Optional[BatchParameters] = None,
    ) -> None:
        self._batch_definition = batch_definition
        self._batch_parameters = batch_parameters
        self.result_format = result_format

        self._get_validator = project_manager.get_validator

    def validate_expectation(
        self,
        expectation: Expectation,
        expectation_parameters: Optional[SuiteParameterDict] = None,
    ) -> ExpectationValidationResult:
        """Run a single expectation against the batch definition"""
        results = self._validate_expectation_configs(
            expectation_configs=[expectation.configuration],
            expectation_parameters=expectation_parameters,
        )

        assert len(results) == 1
        return results[0]

    def validate_expectation_suite(
        self,
        expectation_suite: ExpectationSuite,
        expectation_parameters: Optional[SuiteParameterDict] = None,
    ) -> ExpectationSuiteValidationResult:
        """Run an expectation suite against the batch definition"""
        results = self._validate_expectation_configs(
            expectation_configs=expectation_suite.expectation_configurations,
            expectation_parameters=expectation_parameters,
        )
        statistics = calc_validation_statistics(results)

        return ExpectationSuiteValidationResult(
            results=results,
            success=statistics.success,
            suite_name=expectation_suite.name,
            statistics={
                "evaluated_expectations": statistics.evaluated_expectations,
                "successful_expectations": statistics.successful_expectations,
                "unsuccessful_expectations": statistics.unsuccessful_expectations,
                "success_percent": statistics.success_percent,
            },
            meta={
                # run_id, validation_time, and fkeys are added to this dict
                # in ValidationDefinition.run
                "great_expectations_version": ge_version,
                "batch_spec": convert_to_json_serializable(
                    self._wrapped_validator.active_batch_spec
                ),
                "batch_markers": self._wrapped_validator.active_batch_markers,
                "active_batch_definition": convert_to_json_serializable(
                    self._wrapped_validator.active_batch_definition
                ),
            },
            batch_id=self.active_batch_id,
        )

    @property
    def active_batch_id(self) -> Optional[str]:
        return self._wrapped_validator.active_batch_id

    @property
    def _include_rendered_content(self) -> bool:
        return project_manager.is_using_cloud()

    @cached_property
    def _wrapped_validator(self) -> OldValidator:
        batch_request = self._batch_definition.build_batch_request(
            batch_parameters=self._batch_parameters
        )
        return self._get_validator(batch_request=batch_request)

    def _validate_expectation_configs(
        self,
        expectation_configs: list[ExpectationConfiguration],
        expectation_parameters: Optional[SuiteParameterDict] = None,
    ) -> list[ExpectationValidationResult]:
        """Run a list of expectation configurations against the batch definition"""
        processed_expectation_configs = self._wrapped_validator.process_expectations_for_validation(
            expectation_configs, expectation_parameters
        )

        runtime_configuration: dict
        if isinstance(self.result_format, ResultFormat):
            runtime_configuration = {"result_format": copy(self.result_format.value)}
        else:
            runtime_configuration = {"result_format": copy(self.result_format)}

        results = self._wrapped_validator.graph_validate(
            configurations=processed_expectation_configs,
            runtime_configuration=runtime_configuration,
        )

        if self._include_rendered_content:
            for result in results:
                result.render()

        return results



from __future__ import annotations

from enum import Enum
from typing import Any, Final, List, Literal, Mapping, Union, overload

from great_expectations.compatibility.sqlalchemy import quoted_name
from great_expectations.compatibility.typing_extensions import override


class GXSqlDialect(Enum):
    """Contains sql dialects that have some level of support in Great Expectations.
    Also contains an unsupported attribute if the dialect is not in the list.
    """

    AWSATHENA = "awsathena"
    BIGQUERY = "bigquery"
    DATABRICKS = "databricks"
    DREMIO = "dremio"
    HIVE = "hive"
    MSSQL = "mssql"
    MYSQL = "mysql"
    ORACLE = "oracle"
    POSTGRESQL = "postgresql"
    REDSHIFT = "redshift"
    SNOWFLAKE = "snowflake"
    SQLITE = "sqlite"
    TERADATASQL = "teradatasql"
    TRINO = "trino"
    VERTICA = "vertica"
    CLICKHOUSE = "clickhouse"
    OTHER = "other"

    @override
    def __eq__(self, other: Union[str, bytes, GXSqlDialect]):  # type: ignore[override] # supertype uses `object`
        if isinstance(other, str):
            return self.value.lower() == other.lower()
        # Comparison against byte string, e.g. `b"hive"` should be treated as unicode
        elif isinstance(other, bytes):
            return self.value.lower() == other.lower().decode("utf-8")
        return self.value.lower() == other.value.lower()

    @override
    def __hash__(self: GXSqlDialect):
        return hash(self.value)

    @classmethod
    @override
    def _missing_(cls, value: Any) -> Any:
        try:
            # Sometimes `value` is a byte string, e.g. `b"hive"`, it should be converted
            return cls(value.decode())
        except (UnicodeDecodeError, AttributeError):
            return super()._missing_(value)

    @classmethod
    def get_all_dialect_names(cls) -> List[str]:
        """Get dialect names for all SQL dialects."""
        return [dialect_name.value for dialect_name in cls if dialect_name != GXSqlDialect.OTHER]

    @classmethod
    def get_all_dialects(cls) -> List[GXSqlDialect]:
        """Get all dialects."""
        return [dialect for dialect in cls if dialect != GXSqlDialect.OTHER]


DIALECT_IDENTIFIER_QUOTE_STRINGS: Final[Mapping[GXSqlDialect, Literal['"', "`"]]] = {
    # TODO: add other dialects
    GXSqlDialect.DATABRICKS: "`",
    GXSqlDialect.MYSQL: "`",
    GXSqlDialect.POSTGRESQL: '"',
    GXSqlDialect.SNOWFLAKE: '"',
    GXSqlDialect.SQLITE: '"',
    GXSqlDialect.TRINO: "`",
}


def quote_str(unquoted_identifier: str, dialect: GXSqlDialect) -> str:
    """Quote a string using the specified dialect's quote character."""
    quote_char = DIALECT_IDENTIFIER_QUOTE_STRINGS[dialect]
    if unquoted_identifier.startswith(quote_char) or unquoted_identifier.endswith(quote_char):
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            f"Identifier {unquoted_identifier} already uses quote character {quote_char}"
        )
    return f"{quote_char}{unquoted_identifier}{quote_char}"


def _strip_quotes(s: str, dialect: GXSqlDialect) -> str:
    quote_str = DIALECT_IDENTIFIER_QUOTE_STRINGS[dialect]
    if s.startswith(quote_str) and s.endswith(quote_str):
        return s[1:-1]
    return s


@overload
def wrap_identifier(indentifier: str, dialect: GXSqlDialect = ...) -> quoted_name: ...


@overload
def wrap_identifier(
    indentifier: quoted_name, dialect: GXSqlDialect | None = ...
) -> quoted_name: ...


def wrap_identifier(
    indentifier: str | quoted_name, dialect: GXSqlDialect | None = None
) -> quoted_name:
    if isinstance(indentifier, quoted_name):
        return indentifier
    wo_quotes = _strip_quotes(indentifier, dialect)  # type: ignore[arg-type] # accounted for in overload
    return quoted_name(wo_quotes, quote=True)



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

COL_A = "COL_A"
COL_B = "COL_B"
COL_C = "COL_C"
COL_A_BAD = "COL_A_BAD"
ONES_COL = "ONES_COL"


DATA = pd.DataFrame(
    {
        COL_A: [4, 2, 0],
        COL_B: [2, 7, 7],
        COL_C: [1, -2, 0],
        COL_A_BAD: [4, 4, 0],
        ONES_COL: [1, 1, 1],
    }
)


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_golden_path(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectMulticolumnSumToEqual(column_list=[COL_A, COL_B, COL_C], sum_total=7)
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectMulticolumnSumToEqual(
                column_list=[COL_A_BAD, COL_B, COL_C], sum_total=7, mostly=0.4
            ),
            id="mostly",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(batch_for_datasource: Batch, expectation: gxe.ExpectMulticolumnSumToEqual) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectMulticolumnSumToEqual(
                column_list=[COL_A_BAD, COL_B, COL_C], sum_total=7, mostly=0.7
            ),
            id="mostly_not_met",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(batch_for_datasource: Batch, expectation: gxe.ExpectMulticolumnSumToEqual) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(7, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_sum_total_(
    batch_for_datasource: Batch, suite_param_value: int, expected_result: bool
) -> None:
    suite_param_key = "test_expect_multicolumn_sum_to_equal"
    expectation = gxe.ExpectMulticolumnSumToEqual(
        column_list=[COL_A_BAD, COL_B, COL_C],
        sum_total={"$PARAMETER": suite_param_key},
        mostly=0.4,
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param("all_values_are_missing", True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_ignore_row_if_(
    batch_for_datasource: Batch, suite_param_value: str, expected_result: bool
) -> None:
    suite_param_key = "test_expect_multicolumn_sum_to_equal"
    expectation = gxe.ExpectMulticolumnSumToEqual(
        column_list=[COL_A_BAD, COL_B, COL_C],
        sum_total=7,
        mostly=0.4,
        ignore_row_if={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pandas as pd

from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.sqlalchemy_and_pandas import (
    pandas_read_sql,
)
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.expectations.metrics.table_metric_provider import (
    TableMetricProvider,
)
from great_expectations.validator.metric_configuration import MetricConfiguration
from great_expectations.validator.validator import Validator

if TYPE_CHECKING:
    from great_expectations.compatibility import pyspark


class TableHead(TableMetricProvider):
    metric_name = "table.head"
    value_keys = ("n_rows", "fetch_all")
    default_kwarg_values = {"n_rows": 5, "fetch_all": False}

    @metric_value(engine=PandasExecutionEngine)
    def _pandas(
        cls,
        execution_engine: PandasExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: dict[str, Any],
        runtime_configuration: dict,
    ) -> pd.DataFrame:
        df, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        if metric_value_kwargs.get("fetch_all", cls.default_kwarg_values["fetch_all"]):
            return df
        n_rows: int = (
            metric_value_kwargs.get("n_rows")  # type: ignore[assignment] # FIXME expected 'int', got 'Any | None'
            if metric_value_kwargs.get("n_rows") is not None
            else cls.default_kwarg_values["n_rows"]
        )
        return df.head(n=n_rows)

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: dict[str, Any],
        runtime_configuration: dict,
    ) -> pd.DataFrame:
        selectable, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )

        n_rows: int = (
            metric_value_kwargs.get("n_rows")  # type: ignore[assignment] # FIXME expected 'int', got 'Any | None'
            if metric_value_kwargs.get("n_rows") is not None
            else cls.default_kwarg_values["n_rows"]
        )

        # None means no limit
        limit: int | None = n_rows
        if metric_value_kwargs["fetch_all"]:
            limit = None

        selectable = sa.select("*").select_from(selectable).limit(limit).selectable  # type: ignore[assignment,arg-type] # FIXME CoP

        try:
            with execution_engine.get_connection() as con:
                df = pandas_read_sql(
                    sql=selectable,
                    con=con,
                )
        except StopIteration:
            # empty table. At least try to get the column names
            validator = Validator(execution_engine=execution_engine)
            columns = validator.get_metric(
                MetricConfiguration("table.columns", metric_domain_kwargs)
            )
            df = pd.DataFrame(columns=columns)
        return df  # type: ignore[return-value] # FIXME CoP

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: dict[str, Any],
        runtime_configuration: dict,
    ) -> pd.DataFrame:
        df, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        rows: list[pyspark.Row] | list[dict]
        if metric_value_kwargs["fetch_all"]:
            rows = df.collect()
        else:
            n_rows: int = (
                metric_value_kwargs.get("n_rows")  # type: ignore[assignment] # FIXME expected 'int', got 'Any | None'
                if metric_value_kwargs.get("n_rows") is not None
                else cls.default_kwarg_values["n_rows"]
            )
            if n_rows >= 0:
                rows = df.head(n=n_rows)
            else:
                rows = df.head(n=df.count() + n_rows)

        rows = [element.asDict() for element in rows]
        df = pd.DataFrame(data=rows)  # type: ignore[assignment] # FIXME CoP

        return df  # type: ignore[return-value] # FIXME CoP



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Tuple, Type, Union

from great_expectations.core.metric_function_types import (
    SummarizationMetricNameSuffixes,
)
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    _format_map_output,
    render_suite_parameter_string,
)
from great_expectations.expectations.expectation_configuration import (
    parse_result_format,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import (
    COLUMN_DESCRIPTION,
    MOSTLY_DESCRIPTION,
)
from great_expectations.render import (
    LegacyDiagnosticRendererType,
    LegacyRendererType,
    RenderedStringTemplateContent,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs
    from great_expectations.validator.validator import (
        ValidationDependencies,
    )

EXPECTATION_SHORT_DESCRIPTION = "Expect the column values to be null."
DATA_QUALITY_ISSUES = [DataQualityIssues.COMPLETENESS.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectColumnValuesToBeNull(ColumnMapExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnValuesToBeNull is a \
    Column Map Expectation.

    Column Map Expectations are one of the most common types of Expectation.
    They are evaluated for a single column and ask a yes/no question for every row in that column.
    Based on the result, they then calculate the percentage of rows that gave a positive answer. If the percentage is high enough, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}

    Other Parameters:
        mostly (None or a float between 0 and 1): \
            {MOSTLY_DESCRIPTION} \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly). Default 1.
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    See Also:
        [ExpectColumnValuesToNotBeNull](https://greatexpectations.io/expectations/expect_column_values_to_not_be_null)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	NaN     "A"
            1 	True    NaN
            2 	False   NaN

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnValuesToBeNull(
                    column="test2",
                    mostly=0.66
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 1,
                    "unexpected_percent": 33.33333333333333,
                    "partial_unexpected_list": [
                      "A"
                    ]
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnValuesToBeNull(
                    column="test"
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 2,
                    "unexpected_percent": 66.66666666666666,
                    "partial_unexpected_list": [
                      true,
                      false
                    ]
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    domain_keys: ClassVar[Tuple[str, ...]] = (
        "column",
        "row_condition",
        "condition_parser",
    )

    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "column map expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    map_metric = "column_values.null"
    args_keys = ("column",)

    class Config:
        title = "Expect column values to be null"

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type[ExpectColumnValuesToBeNull]) -> None:
            ColumnMapExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    def _prescriptive_template(
        cls, renderer_configuration: RendererConfiguration
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if params.mostly and params.mostly.value < 1.0:
            renderer_configuration = cls._add_mostly_pct_param(
                renderer_configuration=renderer_configuration
            )
            template_str = "values must be null, at least $mostly_pct % of the time."
        else:
            template_str = "values must be null."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        renderer_configuration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        params = substitute_none_for_missing(
            renderer_configuration.configuration.kwargs,
            ["column", "mostly", "row_condition", "condition_parser"],
        )

        if params["mostly"] is not None:
            if isinstance(params["mostly"], (int, float)) and params["mostly"] < 1.0:
                params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
                # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")  # noqa: E501 # FIXME CoP
            template_str = "values must be null, at least $mostly_pct % of the time."
        else:
            template_str = "values must be null."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        styling = runtime_configuration.get("styling", {}) if runtime_configuration else {}

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    @classmethod
    @renderer(renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE)
    def _diagnostic_observed_value_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        result_dict = result.result

        try:
            notnull_percent = result_dict["unexpected_percent"]
            return num_to_str(100 - notnull_percent, precision=5, use_locale=True) + "% null"
        except KeyError:
            return "unknown % null"
        except TypeError:
            return "NaN% null"

    def get_validation_dependencies(
        self,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> ValidationDependencies:
        validation_dependencies: ValidationDependencies = super().get_validation_dependencies(
            execution_engine, runtime_configuration
        )
        # We do not need this metric for a null metric
        validation_dependencies.remove_metric_configuration(
            metric_name=f"column_values.nonnull.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}"
        )
        return validation_dependencies

    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        result_format = self._get_result_format(runtime_configuration=runtime_configuration)
        mostly = self._get_success_kwargs().get("mostly")
        total_count = metrics.get("table.row_count")
        unexpected_count = metrics.get(
            f"{self.map_metric}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}"
        )

        if total_count is None or total_count == 0:
            # Vacuously true
            success = True
        else:
            success_ratio = (total_count - unexpected_count) / total_count
            success = success_ratio >= mostly

        nonnull_count = None

        return _format_map_output(
            result_format=parse_result_format(result_format),
            success=success,
            element_count=metrics.get("table.row_count"),
            nonnull_count=nonnull_count,
            unexpected_count=metrics.get(
                f"{self.map_metric}.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}"
            ),
            unexpected_list=metrics.get(
                f"{self.map_metric}.{SummarizationMetricNameSuffixes.UNEXPECTED_VALUES.value}"
            ),
            unexpected_index_list=metrics.get(
                f"{self.map_metric}.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}"
            ),
            unexpected_index_query=metrics.get(
                f"{self.map_metric}.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}"
            ),
        )



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import pytest

import great_expectations as gx
import great_expectations.expectations as gxe
from great_expectations.checkpoint import Checkpoint
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.datasource.fluent.interfaces import Datasource

if TYPE_CHECKING:
    from great_expectations.checkpoint.checkpoint import CheckpointResult
    from great_expectations.data_context import AbstractDataContext, FileDataContext

DATASOURCE_NAME = "my_datasource"
ANIMAL_ASSET = "animals_names_asset"
COLUMN_PAIR_ASSET = "column_pair_asset"
MULTI_COLUMN_SUM_ASSET = "multi_column_sum_asset"


@pytest.fixture
def data_context_with_connection_to_metrics_db(
    tmp_path,
) -> FileDataContext:
    """
    Returns DataContext that has a single datasource that connects to a sqlite database.

    The sqlite database (metrics_test.db) contains one table `animal_names` that contains the following data

        "pk_1": [0, 1, 2, 3, 4, 5],
        "pk_2": ["zero", "one", "two", "three", "four", "five"],
        "animals": [
            "cat",
            "fish",
            "dog",
            "giraffe",
            "lion",
            "zebra",
        ],

    It is used by tests for unexpected_index_list (ID/Primary Key).
    """  # noqa: E501 # FIXME CoP

    project_path = tmp_path / "test_configuration"
    context = gx.get_context(mode="file", project_root_dir=project_path)

    sqlite_path = pathlib.Path(__file__).parent / ".." / "test_sets/metrics_test.db"
    assert sqlite_path.exists()

    ds = context.data_sources.add_sqlite(
        name=DATASOURCE_NAME, connection_string=f"sqlite:///{sqlite_path}"
    )
    ds.add_table_asset(name=ANIMAL_ASSET, table_name="animal_names")
    ds.add_table_asset(name=COLUMN_PAIR_ASSET, table_name="column_pairs")
    ds.add_table_asset(name=MULTI_COLUMN_SUM_ASSET, table_name="multi_column_sums")

    project_manager.set_project(context)
    return context


@pytest.fixture
def expected_unexpected_indices_output() -> list[dict]:
    return [
        {"animals": "giraffe", "pk_1": 3},
        {"animals": "lion", "pk_1": 4},
        {"animals": "zebra", "pk_1": 5},
    ]


@pytest.fixture
def expected_sql_query_output() -> str:
    return "SELECT pk_1, animals \n\
FROM animal_names \n\
WHERE animals IS NOT NULL AND (animals NOT IN ('cat', 'fish', 'dog'));"


@pytest.fixture
def expect_column_values_to_be_in_set() -> gxe.ExpectColumnValuesToBeInSet:
    return gxe.ExpectColumnValuesToBeInSet(column="animals", value_set=["cat", "fish", "dog"])


@pytest.fixture
def expect_column_values_to_not_be_in_set() -> gxe.ExpectColumnValuesToNotBeInSet:
    return gxe.ExpectColumnValuesToNotBeInSet(
        column="animals",
        value_set=["giraffe", "lion", "zebra"],
    )


@pytest.fixture
def expect_column_pair_values_to_be_equal() -> gxe.ExpectColumnPairValuesToBeEqual:
    return gxe.ExpectColumnPairValuesToBeEqual(column_A="ordered_item", column_B="received_item")


@pytest.fixture
def expect_multicolumn_sum_to_equal() -> gxe.ExpectMulticolumnSumToEqual:
    return gxe.ExpectMulticolumnSumToEqual(column_list=["a", "b", "c"], sum_total=30)


def _build_checkpoint_and_run(
    context: AbstractDataContext,
    expectations: list[gxe.Expectation],
    asset_name: str,
    result_format: dict,
) -> CheckpointResult:
    suite = context.suites.add(
        suite=ExpectationSuite(name="metrics_exp", expectations=expectations)
    )

    ds = context.data_sources.get("my_datasource")
    assert isinstance(ds, Datasource)
    asset = ds.get_asset(asset_name)
    batch_definition = asset.add_batch_definition(name="my_batch_def")

    validation_definition = context.validation_definitions.add(
        ValidationDefinition(name="my_validation_def", suite=suite, data=batch_definition)
    )

    checkpoint = Checkpoint(
        name="my_checkpoint",
        validation_definitions=[validation_definition],
        result_format=result_format,
    )

    return checkpoint.run()


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_one_expectation_complete_output(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_values_to_be_in_set: gxe.ExpectColumnValuesToBeInSet,
    expected_unexpected_indices_output: list[dict],
    expected_sql_query_output: str,
):
    """
    What does this test?
        - unexpected_index_column defined in Checkpoint only.
        - COMPLETE output, which means we have `unexpected_index_list` and
          `partial_unexpected_index_list`
        - 1 Expectations added to suite
    """
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[expect_column_values_to_be_in_set],
        asset_name=ANIMAL_ASSET,
        result_format={"result_format": "COMPLETE", "unexpected_index_column_names": ["pk_1"]},
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    first_result_full_list = evrs[0]["results"][0]["result"]["unexpected_index_list"]
    assert first_result_full_list == expected_unexpected_indices_output

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == expected_unexpected_indices_output

    unexpected_index_query: str = evrs[0]["results"][0]["result"]["unexpected_index_query"]
    assert unexpected_index_query == expected_sql_query_output


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_one_expectation_complete_output_with_query(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_values_to_be_in_set: gxe.ExpectColumnValuesToBeInSet,
    expected_unexpected_indices_output: list[dict[str, str | int]],
    expected_sql_query_output: str,
):
    """
    What does this test?
        - unexpected_index_column defined in Checkpoint only.
        - COMPLETE output, which means we have `unexpected_index_list` and
         `partial_unexpected_index_list`
        - 1 Expectations added to suite
        - return_unexpected_index_query flag set to True
    """
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[expect_column_values_to_be_in_set],
        asset_name=ANIMAL_ASSET,
        result_format={
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
            "return_unexpected_index_query": True,
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    first_result_full_list = evrs[0]["results"][0]["result"]["unexpected_index_list"]
    assert first_result_full_list == expected_unexpected_indices_output

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == expected_unexpected_indices_output

    unexpected_index_query = evrs[0]["results"][0]["result"]["unexpected_index_query"]
    assert unexpected_index_query == expected_sql_query_output


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_column_pair_expectation_complete_output_with_query(  # noqa: E501 # FIXME CoP
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_pair_values_to_be_equal: gxe.ExpectColumnPairValuesToBeEqual,
):
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[expect_column_pair_values_to_be_equal],
        asset_name=COLUMN_PAIR_ASSET,
        result_format={
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    first_result_full_list = evrs[0]["results"][0]["result"]["unexpected_index_list"]
    assert first_result_full_list == [
        {"pk_1": 3, "ordered_item": "eraser", "received_item": "desk"},
        {"pk_1": 4, "ordered_item": "eraser", "received_item": "desk"},
        {"pk_1": 5, "ordered_item": "eraser", "received_item": "desk"},
    ]

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == [
        {"pk_1": 3, "ordered_item": "eraser", "received_item": "desk"},
        {"pk_1": 4, "ordered_item": "eraser", "received_item": "desk"},
        {"pk_1": 5, "ordered_item": "eraser", "received_item": "desk"},
    ]

    unexpected_index_query = evrs[0]["results"][0]["result"]["unexpected_index_query"]
    assert (
        unexpected_index_query
        == "SELECT pk_1, ordered_item, received_item \nFROM column_pairs \nWHERE NOT (ordered_item = received_item AND NOT (ordered_item IS NULL OR received_item IS NULL));"  # noqa: E501 # FIXME CoP
    )


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_column_pair_expectation_summary_output(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_pair_values_to_be_equal: gxe.ExpectColumnPairValuesToBeEqual,
):
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[expect_column_pair_values_to_be_equal],
        asset_name=COLUMN_PAIR_ASSET,
        result_format={
            "result_format": "SUMMARY",
            "unexpected_index_column_names": ["pk_1"],
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]
    first_result_full_list = evrs[0]["results"][0]["result"].get("unexpected_index_list")
    assert not first_result_full_list

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == [
        {"pk_1": 3, "ordered_item": "eraser", "received_item": "desk"},
        {"pk_1": 4, "ordered_item": "eraser", "received_item": "desk"},
        {"pk_1": 5, "ordered_item": "eraser", "received_item": "desk"},
    ]

    unexpected_index_query = evrs[0]["results"][0]["result"].get("unexpected_index_query")
    assert not unexpected_index_query


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_multi_column_sum_expectation_complete_output_with_query(  # noqa: E501 # FIXME CoP
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_multicolumn_sum_to_equal: gxe.ExpectMulticolumnSumToEqual,
):
    """
    What does this test?
        - unexpected_index_column defined in Checkpoint only.
        - COMPLETE output, which means we have `unexpected_index_list` and
          `partial_unexpected_index_list`
        - 1 Expectations added to suite
        - return_unexpected_index_query flag set to True
    """
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[expect_multicolumn_sum_to_equal],
        asset_name=MULTI_COLUMN_SUM_ASSET,
        result_format={
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
            "return_unexpected_index_query": True,
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    first_result_full_list = evrs[0]["results"][0]["result"]["unexpected_index_list"]
    assert first_result_full_list == [
        {"pk_1": 1, "a": 20, "b": 20, "c": 20},
        {"pk_1": 2, "a": 30, "b": 30, "c": 30},
        {"pk_1": 3, "a": 40, "b": 40, "c": 40},
        {"pk_1": 4, "a": 50, "b": 50, "c": 50},
        {"pk_1": 5, "a": 60, "b": 60, "c": 60},
    ]

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == [
        {"pk_1": 1, "a": 20, "b": 20, "c": 20},
        {"pk_1": 2, "a": 30, "b": 30, "c": 30},
        {"pk_1": 3, "a": 40, "b": 40, "c": 40},
        {"pk_1": 4, "a": 50, "b": 50, "c": 50},
        {"pk_1": 5, "a": 60, "b": 60, "c": 60},
    ]

    unexpected_index_query = evrs[0]["results"][0]["result"]["unexpected_index_query"]
    assert (
        unexpected_index_query
        == "SELECT pk_1, a, b, c \nFROM multi_column_sums \nWHERE 0 + a + b + c != 30.0;"
    )


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_multi_column_sum_expectation_summary_output(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_multicolumn_sum_to_equal: gxe.ExpectMulticolumnSumToEqual,
):
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[expect_multicolumn_sum_to_equal],
        asset_name=MULTI_COLUMN_SUM_ASSET,
        result_format={
            "result_format": "SUMMARY",
            "unexpected_index_column_names": ["pk_1"],
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    first_result_full_list = evrs[0]["results"][0]["result"].get("unexpected_index_list")
    assert not first_result_full_list

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == [
        {"pk_1": 1, "a": 20, "b": 20, "c": 20},
        {"pk_1": 2, "a": 30, "b": 30, "c": 30},
        {"pk_1": 3, "a": 40, "b": 40, "c": 40},
        {"pk_1": 4, "a": 50, "b": 50, "c": 50},
        {"pk_1": 5, "a": 60, "b": 60, "c": 60},
    ]
    unexpected_index_query = evrs[0]["results"][0]["result"].get("unexpected_index_query")
    assert not unexpected_index_query


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_one_expectation_complete_output_no_query(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_values_to_be_in_set: gxe.ExpectColumnValuesToBeInSet,
    expected_unexpected_indices_output: list[dict[str, str | int]],
):
    """
    What does this test?
        - unexpected_index_column defined in Checkpoint only.
        - COMPLETE output, which means we have `unexpected_index_list` and
          `partial_unexpected_index_list`
        - 1 Expectations added to suite
        - return_unexpected_index_query flag set to False
    """
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[expect_column_values_to_be_in_set],
        asset_name=ANIMAL_ASSET,
        result_format={
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
            "return_unexpected_index_query": False,
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    first_result_full_list = evrs[0]["results"][0]["result"]["unexpected_index_list"]
    assert first_result_full_list == expected_unexpected_indices_output

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == expected_unexpected_indices_output

    assert evrs[0]["results"][0]["result"].get("unexpected_index_query") is None


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_two_expectation_complete_output(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_values_to_be_in_set: gxe.ExpectColumnValuesToBeInSet,
    expect_column_values_to_not_be_in_set: gxe.ExpectColumnValuesToNotBeInSet,
    expected_unexpected_indices_output: list[dict[str, str | int]],
    expected_sql_query_output: str,
):
    """
    What does this test?
        - unexpected_index_column not defined in Checkpoint config, but passed in at
          run_checkpoint.
        - COMPLETE output, which means we have `unexpected_index_list` and
          `partial_unexpected_index_list`
        - 2 Expectations added to suite
    """
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[
            expect_column_values_to_be_in_set,
            expect_column_values_to_not_be_in_set,
        ],
        asset_name=ANIMAL_ASSET,
        result_format={
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    # first and second expectations have same results. Although one is "expect_to_be"
    # and the other is "expect_to_not_be", they have opposite value_sets
    first_result_full_list = evrs[0]["results"][0]["result"]["unexpected_index_list"]
    assert first_result_full_list == expected_unexpected_indices_output

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == expected_unexpected_indices_output

    second_result_full_list = evrs[0]["results"][1]["result"]["unexpected_index_list"]
    assert second_result_full_list == expected_unexpected_indices_output

    second_result_partial_list = evrs[0]["results"][1]["result"]["partial_unexpected_index_list"]
    assert second_result_partial_list == expected_unexpected_indices_output

    unexpected_index_query = evrs[0]["results"][0]["result"]["unexpected_index_query"]
    assert unexpected_index_query == expected_sql_query_output


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_one_expectation_summary_output(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_values_to_be_in_set: gxe.ExpectColumnValuesToBeInSet,
    expected_unexpected_indices_output: list[dict[str, str | int]],
):
    """
    What does this test?
        - unexpected_index_column defined in Checkpoint only.
        - SUMMARY output, which means we have `partial_unexpected_index_list` only
        - 1 Expectations added to suite
    """
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[
            expect_column_values_to_be_in_set,
        ],
        asset_name=ANIMAL_ASSET,
        result_format={
            "result_format": "SUMMARY",
            "unexpected_index_column_names": ["pk_1"],
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    first_result_full_list = evrs[0]["results"][0]["result"].get("unexpected_index_list")
    assert not first_result_full_list

    first_result_partial_list = evrs[0]["results"][0]["result"]["partial_unexpected_index_list"]
    assert first_result_partial_list == expected_unexpected_indices_output


@pytest.mark.filesystem
def test_sql_result_format_in_checkpoint_pk_defined_one_expectation_basic_output(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_values_to_be_in_set: gxe.ExpectColumnValuesToBeInSet,
):
    """
    What does this test?
        - unexpected_index_column defined in Checkpoint only.
        - BASIC output, which means we have no unexpected_index_list output
        - 1 Expectations added to suite
    """
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[
            expect_column_values_to_be_in_set,
        ],
        asset_name=ANIMAL_ASSET,
        result_format={
            "result_format": "BASIC",
            "unexpected_index_column_names": ["pk_1"],
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"]["unexpected_index_column_names"]
    assert index_column_names == ["pk_1"]

    first_result_full_list = evrs[0]["results"][0]["result"].get("unexpected_index_list")
    assert not first_result_full_list

    first_result_partial_list = evrs[0]["results"][0]["result"].get("partial_unexpected_index_list")
    assert not first_result_partial_list

    assert evrs[0]["results"][0]["result"].get("unexpected_index_query") is None


@pytest.mark.filesystem
def test_sql_complete_output_no_id_pk_fallback(
    data_context_with_connection_to_metrics_db: FileDataContext,
    expect_column_values_to_be_in_set: gxe.ExpectColumnValuesToBeInSet,
):
    result = _build_checkpoint_and_run(
        context=data_context_with_connection_to_metrics_db,
        expectations=[
            expect_column_values_to_be_in_set,
        ],
        asset_name=ANIMAL_ASSET,
        result_format={
            "result_format": "COMPLETE",
        },
    )
    evrs = list(result.run_results.values())

    index_column_names = evrs[0]["results"][0]["result"].get("unexpected_index_column_names")
    assert not index_column_names

    first_result_full_list = evrs[0]["results"][0]["result"].get("unexpected_index_list")
    assert not first_result_full_list

    first_result_partial_list = evrs[0]["results"][0]["result"].get("partial_unexpected_index_list")
    assert not first_result_partial_list

    unexpected_index_query = evrs[0]["results"][0]["result"].get("unexpected_index_query")
    # query does not contain id_pk column
    assert (
        unexpected_index_query
        == "SELECT animals \nFROM animal_names \nWHERE animals IS NOT NULL AND (animals NOT IN ('cat', 'fish', 'dog'));"  # noqa: E501 # FIXME CoP
    )



from typing import Sequence

import pandas as pd

from great_expectations.expectations.metrics.util import MAX_RESULT_RECORDS
from great_expectations.metrics import QueryDataSourceTable
from great_expectations.metrics.query.data_source_table import QueryDataSourceTableResult
from tests.integration.conftest import (
    MultiSourceBatch,
    MultiSourceTestConfig,
    multi_source_batch_setup,
)
from tests.integration.test_utils.data_source_config import (
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

SPARK_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    SparkFilesystemCsvDatasourceTestConfig(),
]

BASE_DATA_FRAME = pd.DataFrame({"foo": [1, 2, 3], "bar": [4, 5, 6]})

COMPARISON_DATA_FRAME = pd.DataFrame(
    {
        "id": [1, 2, 3, 4],
        "name": ["A", "B", "C", "A"],
    },
)

BIG_COMPARISON_DATA_FRAME = pd.DataFrame(
    {
        "id": [i for i in range(300)],
        "name": ["A" for _ in range(300)],
    }
)

ALL_COMPARISON_TO_BASE_SOURCES = [
    MultiSourceTestConfig(
        comparison=PostgreSQLDatasourceTestConfig(), base=PostgreSQLDatasourceTestConfig()
    ),
    MultiSourceTestConfig(
        comparison=PostgreSQLDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=SnowflakeDatasourceTestConfig(), base=SnowflakeDatasourceTestConfig()
    ),
    MultiSourceTestConfig(
        comparison=SnowflakeDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=DatabricksDatasourceTestConfig(),
        base=DatabricksDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=DatabricksDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=RedshiftDatasourceTestConfig(),
        base=RedshiftDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=RedshiftDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    ),
]


class TestQueryRowCount:
    @multi_source_batch_setup(
        multi_source_test_configs=ALL_COMPARISON_TO_BASE_SOURCES,
        base_data=BASE_DATA_FRAME,
        comparison_data=COMPARISON_DATA_FRAME,
    )
    def test_success_sql(self, multi_source_batch: MultiSourceBatch) -> None:
        query = f"SELECT * FROM {multi_source_batch.comparison_table_name} WHERE name = 'A';"
        batch = multi_source_batch.base_batch
        metric = QueryDataSourceTable(
            query=query, data_source_name=multi_source_batch.comparison_data_source_name
        )
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, QueryDataSourceTableResult)
        assert len(metric_result.value) == 2

    @multi_source_batch_setup(
        multi_source_test_configs=ALL_COMPARISON_TO_BASE_SOURCES,
        base_data=BASE_DATA_FRAME,
        comparison_data=BIG_COMPARISON_DATA_FRAME,
    )
    def test_result_is_limited_to_200_rows(self, multi_source_batch: MultiSourceBatch) -> None:
        query = f"SELECT * FROM {multi_source_batch.comparison_table_name} WHERE id > 0"
        batch = multi_source_batch.base_batch
        metric = QueryDataSourceTable(
            query=query, data_source_name=multi_source_batch.comparison_data_source_name
        )
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, QueryDataSourceTableResult)
        assert len(metric_result.value) == MAX_RESULT_RECORDS



from __future__ import annotations

from enum import Enum


class FontFamily(Enum):
    MONTSERRAT = "Montserrat"
    ROBOTO_MONO = "Roboto Mono"
    SOURCE_SANS_PRO = "Source Sans Pro"


class FontFamilyURL(Enum):
    MONTSERRAT = "https://fonts.googleapis.com/css2?family=Montserrat"
    ROBOTO_MONO = "https://fonts.googleapis.com/css2?family=Roboto+Mono"
    SOURCE_SANS_PRO = "https://fonts.googleapis.com/css2?family=Source+Sans+Pro"



import datetime
import os
from typing import List, Tuple
from unittest import mock

import numpy as np
import pandas as pd
import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility import pyarrow, pyspark
from great_expectations.compatibility.pyspark import functions as F
from great_expectations.core.batch_spec import (
    AzureBatchSpec,
    GCSBatchSpec,
    PathBatchSpec,
    RuntimeDataBatchSpec,
    S3BatchSpec,
)
from great_expectations.execution_engine.partition_and_sample.data_partitioner import (
    DatePart,
)
from great_expectations.execution_engine.partition_and_sample.sparkdf_data_partitioner import (
    SparkDataPartitioner,
)
from tests.execution_engine.partition_and_sample.partition_and_sample_test_cases import (
    MULTIPLE_DATE_PART_BATCH_IDENTIFIERS,
    MULTIPLE_DATE_PART_DATE_PARTS,
    SINGLE_DATE_PART_BATCH_IDENTIFIERS,
    SINGLE_DATE_PART_DATE_PARTS,
)

# Here we add SparkDataPartitioner specific test cases to the generic test cases:
SINGLE_DATE_PART_DATE_PARTS += [
    pytest.param(
        [SparkDataPartitioner.date_part.MONTH],
        id="month getting date parts from SparkDataPartitioner.date_part",
    )
]
MULTIPLE_DATE_PART_DATE_PARTS += [
    pytest.param(
        [SparkDataPartitioner.date_part.YEAR, SparkDataPartitioner.date_part.MONTH],
        id="year_month getting date parts from SparkDataPartitioner.date_part",
    )
]

# module level markers
pytestmark = pytest.mark.spark


@pytest.fixture
def simple_multi_year_spark_df(spark_session):
    spark_df_data: List[Tuple] = [
        ("2018-01-01 12:00:00.000",),
        ("2018-10-02 12:00:00.000",),
        ("2019-01-01 12:00:00.000",),
        ("2019-10-02 12:00:00.000",),
        ("2019-11-03 12:00:00.000",),
        ("2020-01-01 12:00:00.000",),
        ("2020-10-02 12:00:00.000",),
        ("2020-11-03 12:00:00.000",),
        ("2020-12-04 12:00:00.000",),
    ]

    spark_df: pyspark.DataFrame = spark_session.createDataFrame(
        data=spark_df_data, schema=["input_timestamp"]
    )
    spark_df = spark_df.withColumn("timestamp", F.to_timestamp("input_timestamp"))
    assert spark_df.count() == 9
    return spark_df


@pytest.mark.parametrize(
    "partitioner_kwargs_year,num_values_in_df",
    [
        pytest.param(year, num_values, id=year)
        for year, num_values in {"2018": 2, "2019": 3, "2020": 4}.items()
    ],
)
def test_get_batch_with_partition_on_year(
    partitioner_kwargs_year,
    num_values_in_df,
    spark_session,
    basic_spark_df_execution_engine,
    simple_multi_year_spark_df: pyspark.DataFrame,
):
    partitioned_df: pyspark.DataFrame = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=simple_multi_year_spark_df,
            partitioner_method="partition_on_year",
            partitioner_kwargs={
                "column_name": "timestamp",
                "batch_identifiers": {"timestamp": partitioner_kwargs_year},
            },
        )
    ).dataframe
    assert partitioned_df.count() == num_values_in_df
    assert len(partitioned_df.columns) == 2


@pytest.mark.parametrize(
    "column_batch_identifier,num_values_in_df",
    [
        pytest.param(column_batch_identifier, num_values, id=column_batch_identifier)
        for column_batch_identifier, num_values in {
            "2018-01-01": 3,
            "2018-01-02": 3,
            "2018-01-03": 2,
            "2018-01-04": 1,
        }.items()
    ],
)
def test_get_batch_with_partition_on_date_parts_day(
    column_batch_identifier,
    num_values_in_df,
    spark_session,
    basic_spark_df_execution_engine,
    simple_multi_year_spark_df: pyspark.DataFrame,
):
    partitioned_df: pyspark.DataFrame = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=simple_multi_year_spark_df,
            partitioner_method="partition_on_date_parts",
            partitioner_kwargs={
                "column_name": "timestamp",
                "batch_identifiers": {"timestamp": column_batch_identifier},
                "date_parts": [DatePart.DAY],
            },
        )
    ).dataframe

    assert partitioned_df.count() == num_values_in_df
    assert len(partitioned_df.columns) == 2


@pytest.mark.parametrize(
    "batch_identifiers_for_column",
    SINGLE_DATE_PART_BATCH_IDENTIFIERS,
)
@pytest.mark.parametrize(
    "date_parts",
    SINGLE_DATE_PART_DATE_PARTS,
)
def test_partition_on_date_parts_single_date_parts(
    batch_identifiers_for_column, date_parts, simple_multi_year_spark_df
):
    """What does this test and why?

    partition_on_date_parts should still filter the correct rows from the input dataframe when passed a single element list
     date_parts that is a string, DatePart enum objects, mixed case string.
     To match our interface it should accept a dateutil parseable string as the batch identifier
     or a datetime and also fail when parameters are invalid.
    """  # noqa: E501 # FIXME CoP
    data_partitioner: SparkDataPartitioner = SparkDataPartitioner()
    column_name: str = "timestamp"
    result: pyspark.DataFrame = data_partitioner.partition_on_date_parts(
        df=simple_multi_year_spark_df,
        column_name=column_name,
        batch_identifiers={column_name: batch_identifiers_for_column},
        date_parts=date_parts,
    )
    assert result.count() == 3


@pytest.mark.parametrize(
    "batch_identifiers_for_column",
    MULTIPLE_DATE_PART_BATCH_IDENTIFIERS,
)
@pytest.mark.parametrize(
    "date_parts",
    MULTIPLE_DATE_PART_DATE_PARTS,
)
def test_partition_on_date_parts_multiple_date_parts(
    batch_identifiers_for_column, date_parts, simple_multi_year_spark_df
):
    """What does this test and why?

    partition_on_date_parts should still filter the correct rows from the input dataframe when passed
     date parts that are strings, DatePart enum objects, a mixture and mixed case.
     To match our interface it should accept a dateutil parseable string as the batch identifier
     or a datetime and also fail when parameters are invalid.
    """  # noqa: E501 # FIXME CoP
    data_partitioner: SparkDataPartitioner = SparkDataPartitioner()
    column_name: str = "timestamp"
    result: pyspark.DataFrame = data_partitioner.partition_on_date_parts(
        df=simple_multi_year_spark_df,
        column_name=column_name,
        batch_identifiers={column_name: batch_identifiers_for_column},
        date_parts=date_parts,
    )
    assert result.count() == 1


@mock.patch(
    "great_expectations.execution_engine.partition_and_sample.sparkdf_data_partitioner.SparkDataPartitioner.partition_on_date_parts"
)
@pytest.mark.parametrize(
    "partitioner_method_name,called_with_date_parts",
    [
        ("partition_on_year", [DatePart.YEAR]),
        ("partition_on_year_and_month", [DatePart.YEAR, DatePart.MONTH]),
        (
            "partition_on_year_and_month_and_day",
            [DatePart.YEAR, DatePart.MONTH, DatePart.DAY],
        ),
    ],
)
def test_named_date_part_methods(
    mock_partition_on_date_parts: mock.MagicMock,  # noqa: TID251 # FIXME CoP
    partitioner_method_name: str,
    called_with_date_parts: List[DatePart],
    simple_multi_year_spark_df: pyspark.DataFrame,
):
    """Test that a partially pre-filled version of partition_on_date_parts() was called with the appropriate params.
    For example, partition_on_year.
    """  # noqa: E501 # FIXME CoP
    data_partitioner: SparkDataPartitioner = SparkDataPartitioner()
    column_name: str = "column_name"
    batch_identifiers: dict = {column_name: {"year": 2018, "month": 10, "day": 31}}

    getattr(data_partitioner, partitioner_method_name)(
        df=simple_multi_year_spark_df,
        column_name=column_name,
        batch_identifiers=batch_identifiers,
    )

    mock_partition_on_date_parts.assert_called_with(
        df=simple_multi_year_spark_df,
        column_name=column_name,
        batch_identifiers=batch_identifiers,
        date_parts=called_with_date_parts,
    )


@pytest.mark.parametrize(
    "underscore_prefix",
    [
        pytest.param("_", id="underscore prefix"),
        pytest.param("", id="no underscore prefix"),
    ],
)
@pytest.mark.parametrize(
    "partitioner_method_name",
    [
        pytest.param(partitioner_method_name, id=partitioner_method_name)
        for partitioner_method_name in [
            "partition_on_year",
            "partition_on_year_and_month",
            "partition_on_year_and_month_and_day",
            "partition_on_date_parts",
            "partition_on_whole_table",
            "partition_on_column_value",
            "partition_on_converted_datetime",
            "partition_on_divided_integer",
            "partition_on_mod_integer",
            "partition_on_multi_column_values",
            "partition_on_hashed_column",
        ]
    ],
)
def test_get_partitioner_method(underscore_prefix: str, partitioner_method_name: str):
    data_partitioner: SparkDataPartitioner = SparkDataPartitioner()

    partitioner_method_name_with_prefix = f"{underscore_prefix}{partitioner_method_name}"

    assert data_partitioner.get_partitioner_method(partitioner_method_name_with_prefix) == getattr(
        data_partitioner, partitioner_method_name
    )


def test_get_batch_empty_partitioner(
    test_folder_connection_path_csv, basic_spark_df_execution_engine
):
    # reader_method not configured because spark will configure own reader by default
    # reader_options are needed to specify the fact that the first line of test file is the header
    test_sparkdf = basic_spark_df_execution_engine.get_batch_data(
        PathBatchSpec(
            path=os.path.join(  # noqa: PTH118 # FIXME CoP
                test_folder_connection_path_csv, "test.csv"
            ),
            reader_options={"header": True},
            partitioner_method=None,
        )
    ).dataframe
    assert test_sparkdf.count() == 5
    assert len(test_sparkdf.columns) == 2


def test_get_batch_empty_partitioner_tsv(
    test_folder_connection_path_tsv, basic_spark_df_execution_engine
):
    # reader_method not configured because spark will configure own reader by default
    # reader_options are needed to specify the fact that the first line of test file is the header
    # reader_options are also needed to specify the separator (otherwise, comma will be used as the default separator)  # noqa: E501 # FIXME CoP
    test_sparkdf = basic_spark_df_execution_engine.get_batch_data(
        PathBatchSpec(
            path=os.path.join(  # noqa: PTH118 # FIXME CoP
                test_folder_connection_path_tsv, "test.tsv"
            ),
            reader_options={"header": True, "sep": "\t"},
            partitioner_method=None,
        )
    ).dataframe
    assert test_sparkdf.count() == 5
    assert len(test_sparkdf.columns) == 2


@pytest.mark.skipif(
    not pyarrow.pyarrow,
    reason='Could not import "pyarrow"',
)
def test_get_batch_empty_partitioner_parquet(
    test_folder_connection_path_parquet, basic_spark_df_execution_engine
):
    # Note: reader method and reader_options are not needed, because
    # SparkDFExecutionEngine automatically determines the file type as well as the schema of the Parquet file.  # noqa: E501 # FIXME CoP
    test_sparkdf = basic_spark_df_execution_engine.get_batch_data(
        PathBatchSpec(
            path=os.path.join(  # noqa: PTH118 # FIXME CoP
                test_folder_connection_path_parquet, "test.parquet"
            ),
            partitioner_method=None,
        )
    ).dataframe
    assert test_sparkdf.count() == 5
    assert len(test_sparkdf.columns) == 2


def test_get_batch_with_partition_on_whole_table_runtime(
    test_sparkdf, basic_spark_df_execution_engine
):
    test_sparkdf = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf, partitioner_method="_partition_on_whole_table"
        )
    ).dataframe
    assert test_sparkdf.count() == 120
    assert len(test_sparkdf.columns) == 10


def test_get_batch_with_partition_on_whole_table_filesystem(
    test_folder_connection_path_csv, basic_spark_df_execution_engine
):
    # reader_method not configured because spark will configure own reader by default
    test_sparkdf = basic_spark_df_execution_engine.get_batch_data(
        PathBatchSpec(
            path=os.path.join(  # noqa: PTH118 # FIXME CoP
                test_folder_connection_path_csv, "test.csv"
            ),
            partitioner_method="_partition_on_whole_table",
        )
    ).dataframe
    assert test_sparkdf.count() == 6
    assert len(test_sparkdf.columns) == 2


def test_get_batch_with_partition_on_whole_table_s3(spark_session, basic_spark_df_execution_engine):
    # noinspection PyUnusedLocal
    def mocked_get_reader_function(*args, **kwargs):
        # noinspection PyUnusedLocal,PyShadowingNames
        def mocked_reader_function(*args, **kwargs):
            pd_df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [2, 3, 4, None]})
            df = spark_session.createDataFrame(
                [
                    tuple(
                        None if isinstance(x, (float, int)) and np.isnan(x) else x
                        for x in record.tolist()
                    )
                    for record in pd_df.to_records(index=False)
                ],
                pd_df.columns.tolist(),
            )
            return df

        return mocked_reader_function

    spark_engine = basic_spark_df_execution_engine
    spark_engine._get_reader_fn = mocked_get_reader_function

    test_sparkdf = spark_engine.get_batch_data(
        S3BatchSpec(
            path="s3://bucket/test/test.csv",
            reader_method="csv",
            reader_options={"header": True},
            partitioner_method="_partition_on_whole_table",
        )
    ).dataframe
    assert test_sparkdf.count() == 4
    assert len(test_sparkdf.columns) == 2


def test_get_batch_with_partition_on_whole_table_azure(
    spark_session, basic_spark_df_execution_engine
):
    # noinspection PyUnusedLocal
    def mocked_get_reader_function(*args, **kwargs):
        # noinspection PyUnusedLocal,PyShadowingNames
        def mocked_reader_function(*args, **kwargs):
            pd_df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [2, 3, 4, None]})
            df = spark_session.createDataFrame(
                [
                    tuple(
                        None if isinstance(x, (float, int)) and np.isnan(x) else x
                        for x in record.tolist()
                    )
                    for record in pd_df.to_records(index=False)
                ],
                pd_df.columns.tolist(),
            )
            return df

        return mocked_reader_function

    spark_engine = basic_spark_df_execution_engine
    spark_engine._get_reader_fn = mocked_get_reader_function

    test_sparkdf = spark_engine.get_batch_data(
        AzureBatchSpec(
            path="wasbs://test_container@test_account.blob.core.windows.net/test_dir/test_file.csv",
            reader_method="csv",
            reader_options={"header": True},
            partitioner_method="_partition_on_whole_table",
        )
    ).dataframe
    assert test_sparkdf.count() == 4
    assert len(test_sparkdf.columns) == 2


def test_get_batch_with_partition_on_whole_table_gcs(
    spark_session, basic_spark_df_execution_engine
):
    # noinspection PyUnusedLocal
    def mocked_get_reader_function(*args, **kwargs):
        # noinspection PyUnusedLocal,PyShadowingNames
        def mocked_reader_function(*args, **kwargs):
            pd_df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [2, 3, 4, None]})
            df = spark_session.createDataFrame(
                [
                    tuple(
                        None if isinstance(x, (float, int)) and np.isnan(x) else x
                        for x in record.tolist()
                    )
                    for record in pd_df.to_records(index=False)
                ],
                pd_df.columns.tolist(),
            )
            return df

        return mocked_reader_function

    spark_engine = basic_spark_df_execution_engine
    spark_engine._get_reader_fn = mocked_get_reader_function

    test_sparkdf = spark_engine.get_batch_data(
        GCSBatchSpec(
            path="gcs://bucket/test/test.csv",
            reader_method="csv",
            reader_options={"header": True},
            partitioner_method="_partition_on_whole_table",
        )
    ).dataframe
    assert test_sparkdf.count() == 4
    assert len(test_sparkdf.columns) == 2


def test_get_batch_with_partition_on_column_value(test_sparkdf, basic_spark_df_execution_engine):
    partitioned_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            partitioner_method="_partition_on_column_value",
            partitioner_kwargs={
                "column_name": "batch_id",
                "batch_identifiers": {"batch_id": 2},
            },
        )
    ).dataframe
    assert test_sparkdf.count() == 120
    assert len(test_sparkdf.columns) == 10
    collected = partitioned_df.collect()
    for val in collected:
        assert val.batch_id == 2

    partitioned_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            partitioner_method="_partition_on_column_value",
            partitioner_kwargs={
                "column_name": "date",
                "batch_identifiers": {"date": datetime.date(2020, 1, 30)},
            },
        )
    ).dataframe
    assert partitioned_df.count() == 3
    assert len(partitioned_df.columns) == 10


def test_get_batch_with_partition_on_converted_datetime(
    test_sparkdf, basic_spark_df_execution_engine
):
    partitioned_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            partitioner_method="_partition_on_converted_datetime",
            partitioner_kwargs={
                "column_name": "timestamp",
                "batch_identifiers": {"timestamp": "2020-01-03"},
            },
        )
    ).dataframe
    assert partitioned_df.count() == 2
    assert len(partitioned_df.columns) == 10


def test_get_batch_with_partition_on_divided_integer(test_sparkdf, basic_spark_df_execution_engine):
    partitioned_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            partitioner_method="_partition_on_divided_integer",
            partitioner_kwargs={
                "column_name": "id",
                "divisor": 10,
                "batch_identifiers": {"id": 5},
            },
        )
    ).dataframe
    assert partitioned_df.count() == 10
    assert len(partitioned_df.columns) == 10
    max_result = partitioned_df.select([F.max("id")])
    assert max_result.collect()[0]["max(id)"] == 59
    min_result = partitioned_df.select([F.min("id")])
    assert min_result.collect()[0]["min(id)"] == 50


def test_get_batch_with_partition_on_mod_integer(test_sparkdf, basic_spark_df_execution_engine):
    partitioned_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            partitioner_method="_partition_on_mod_integer",
            partitioner_kwargs={
                "column_name": "id",
                "mod": 10,
                "batch_identifiers": {"id": 5},
            },
        )
    ).dataframe

    assert partitioned_df.count() == 12
    assert len(partitioned_df.columns) == 10
    max_result = partitioned_df.select([F.max("id")])
    assert max_result.collect()[0]["max(id)"] == 115
    min_result = partitioned_df.select([F.min("id")])
    assert min_result.collect()[0]["min(id)"] == 5


def test_get_batch_with_partition_on_multi_column_values(
    test_sparkdf, basic_spark_df_execution_engine
):
    partitioned_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            partitioner_method="_partition_on_multi_column_values",
            partitioner_kwargs={
                "column_names": ["y", "m", "d"],
                "batch_identifiers": {
                    "y": 2020,
                    "m": 1,
                    "d": 5,
                },
            },
        )
    ).dataframe
    assert partitioned_df.count() == 4
    assert len(partitioned_df.columns) == 10
    collected = partitioned_df.collect()
    for val in collected:
        assert val.date == datetime.date(2020, 1, 5)

    with pytest.raises(ValueError):
        # noinspection PyUnusedLocal
        partitioned_df = basic_spark_df_execution_engine.get_batch_data(
            RuntimeDataBatchSpec(
                batch_data=test_sparkdf,
                partitioner_method="_partition_on_multi_column_values",
                partitioner_kwargs={
                    "column_names": ["I", "dont", "exist"],
                    "batch_identifiers": {
                        "y": 2020,
                        "m": 1,
                        "d": 5,
                    },
                },
            )
        ).dataframe


def test_get_batch_with_partition_on_hashed_column_incorrect_hash_function_name(
    test_sparkdf,
    basic_spark_df_execution_engine,
):
    with pytest.raises(gx_exceptions.ExecutionEngineError):
        # noinspection PyUnusedLocal
        _ = basic_spark_df_execution_engine.get_batch_data(
            RuntimeDataBatchSpec(
                batch_data=test_sparkdf,
                partitioner_method="_partition_on_hashed_column",
                partitioner_kwargs={
                    "column_name": "favorite_color",
                    "hash_digits": 1,
                    "hash_function_name": "I_wont_work",
                    "batch_identifiers": {
                        "hash_value": "a",
                    },
                },
            )
        ).dataframe


def test_get_batch_with_partition_on_hashed_column(test_sparkdf, basic_spark_df_execution_engine):
    partitioned_df = basic_spark_df_execution_engine.get_batch_data(
        RuntimeDataBatchSpec(
            batch_data=test_sparkdf,
            partitioner_method="_partition_on_hashed_column",
            partitioner_kwargs={
                "column_name": "favorite_color",
                "hash_digits": 1,
                "hash_function_name": "sha256",
                "batch_identifiers": {
                    "hash_value": "a",
                },
            },
        )
    ).dataframe
    assert partitioned_df.count() == 8
    assert len(partitioned_df.columns) == 10
